#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A collection of utility functions for plotting"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import copy
import gc

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap

from .. pygeostat_parameters import Parameters

def titleoverlap(ax, title):
    """
    A small utility function called from many of the plotting functions. This returns a boolean if
    the title overlaps with x-axis tick labels that have been placed on the
    top spline.

    Parameters:
        ax (ax): A matplotlib axis object with x axis tick labels
        title (text object): A text object created by plt.title(), plt.suptitle(), or plt.text()
            (e.g., titletxt = plt.suptitle('Title', y=0.98))

    Returns:
        overlap (bool): Indicator if the x-axis tick labels overlap the plot title.

    .. codeauthor:: pygeostat development team - 2015-10-07
    """
    # Set-up a function that completes the check
    def ticktitlecheck(labels, title):
        """Check to see if the x-axis tick labels overlap the plot title"""
        fig = plt.gcf()
        renderer = fig.canvas.get_renderer()
        fig.draw(renderer)
        if not labels or not title:
            return False
        else:
            try:
                xboxes = [label.get_window_extent() for label in labels]
                titlebox = title.get_window_extent()
                overlaps = titlebox.count_overlaps(xboxes)
                return overlaps >= 1
            except:
                print("There was an error checking for overlapping x axis tick labels and plot",
                      "title. You'll have to set the argument manually to rotate them if necessary")
    overlaps = ticktitlecheck(ax.get_xticklabels(), title)

    return overlaps


def supaxislabel(axis, label, rotation=None, label_prop=None, labelpad=0, ha='center', va='center',
                 fig=None):
    """
    A small utility function called from many of the plotting functions. This adds super ylabel or
    xlabel to the figure similar to mpl.suptitle when using subplots.
    From: http://stackoverflow.com/a/29107972

    Parameters:
        axis (str): Indicator as to which axis to place a label. Allows 'x' or 'y' as parameters
        label (str): Label to place on the selected axis
        label_prop: Keyword dictionary to be set to plt.text
        labelpad (float): Padding from the axis (default: 0)
        ha (str): horizontal alignment (default: 'center')
        va (str): vertical alignment (default: 'center')

    .. codeauthor:: pygeostat development team - 2015-09-30
    """
    if fig is None:
        fig = plt.gcf()
    xmin = []
    ymin = []
    for ax in fig.axes:
        xmin.append(ax.get_position().xmin)
        ymin.append(ax.get_position().ymin)
    xmin, ymin = min(xmin), min(ymin)
    dpi = fig.dpi
    if axis.lower() == "y":
        if rotation is None:
            rotation = 90
        x = xmin - float(labelpad) / dpi
        y = 0.5
    elif axis.lower() == 'x':
        if rotation is None:
            rotation = 0
        x = 0.5
        y = ymin - float(labelpad) / dpi
    else:
        raise Exception("Unexpected axis: x or y")
    if label_prop is None:
        label_prop = dict()
    plt.text(x, y, label, rotation=rotation, transform=fig.transFigure, ha=ha, va=va, **label_prop)


def getminmax(vlim, sigfigs):
    """
    A small utility function called from many of the plotting functions. This determines some
    logical minimum and maximum values to bound a dataset based on a specified number of
    significant figures.

    Parameters:
        vlim (float tuple): Minimum and maximum values to honor if  parameter of "None" is not
            passed (vmin, vmax)
        sigfigs (int): Number of significant figures to consider

    Returns:
        valrng (float tuple): Calculated logical (min, max) bounds

    .. codeauthor:: pygeostat development team 2015-10-13
    """
    # Set-up some parameters
    valrng = [copy.deepcopy(vlim[0]), copy.deepcopy(vlim[1])]
    valabs = [False, False]
    flip = False
    # Get absolute values if required
    if valrng[0] < 0:
        valabs[0] = True
        valrng[0] = abs(valrng[0])
    if valrng[1] < 0:
        valabs[1] = True
        valrng[1] = abs(valrng[1])
    # Flip the value range if required
    if valrng[0] > valrng[1]:
        flip = True
        valrng = valrng[::-1]
    # Get logical min/max values
    try:
        precision = int(('{:.%ie}' % sigfigs).format(valrng[1]).partition('+')[2])
    except ValueError:
        precision = -1 * int(('{:.%ie}' % sigfigs).format(valrng[1]).partition('-')[2])
    valrng[1] = valrng[1] / (10**(precision - sigfigs + 1))
    valrng[0] = valrng[0] / (10**(precision - sigfigs + 1))
    valrng[1] = valrng[1] + 0.5
    if valabs[0]:
        valrng[0] = valrng[0] + 0.5
    else:
        valrng[0] = valrng[0] - 0.5
    valrng[1] = round(valrng[1], 0) * (10.0**(precision - sigfigs + 1)) + 0
    valrng[0] = round(valrng[0], 0) * (10.0**(precision - sigfigs + 1)) + 0
    # Flip back if required
    if flip:
        valrng = valrng[::-1]
    # Undo absolute values if required
    if valabs[0]:
        valrng[0] = valrng[0] * -1 + 0
    if valabs[1]:
        valrng[1] = valrng[1] * -1 + 0

    return (valrng[0], valrng[1])


def get_contcbarargs(data, sigfigs, vlim=None, nticks=5, catdata=None):
    """
    A small utility function called from many of the plotting functions. This determines logical
    continuous colorbar arguments if value limit values are not provided.
    If they are, determine some intermediate values.

    Parameters:
        data: A numpy array containing the data that is being plotted and required a colorbar
        sigfigs (int): The number of significant figures to consider
        vlim (float tuple): Value limits of the data being plotted as a tuple (vmin, vmax)
        nticks (int): The number of tick labels that need to be generated
        catdata (bool): Basically assume the codes are integers

    Returns
    -------
        vlim: float tuple
            A tuple of the new limits to use for plotting
        ticklocs: list
            A list of tick locations for the colorbar
        ticklabels: list
            A list of tick labels for each tickloc value

    .. codeauthor:: pygeostat development team 2015-09-30
    """
    vmin = np.nanmin(np.array(data))
    vmax = np.nanmax(np.array(data))
    # If limits weren't provided, get them from the data and set-up a boolean for later
    if vlim is None:
        vlim = (vmin, vmax)
        # Move the vmin and vmax values to something logical
        vlim = getminmax(vlim, sigfigs)
        vlims = False
    else:
        tvlim = []
        if hasattr(vlim[0], 'values'):
            tvlim.append(vlim[0].values[0])
        else:
            tvlim.append(vmin if vlim[0] is None else vlim[0])
        if hasattr(vlim[1], 'values'):
            tvlim.append(vlim[1].values[0])
        else:
            tvlim.append(vmax if vlim[1] is None else vlim[1])
        vlim = tuple(tvlim)
        vlims = True
    # Determine the other tick locations from highest to lowest
    step = float(1 / (nticks - 1))  # step percentage of each tick
    diff = abs(vlim[1] - vlim[0])  # difference between the limits
    if vlims:
        ticklocs = [vlim[1]]
    else:
        ticklocs = [float(('{:.%ig}' % sigfigs).format(vlim[1]))]
    for i in range(nticks - 2, 0, -1):
        value = diff * (i * step) + vlim[0]
        value = float(('{:.%ig}' % sigfigs).format(value))
        ticklocs.append(value)
    if vlims:
        ticklocs.append(vlim[0])
    else:
        ticklocs.append(float(('{:.%ig}' % sigfigs).format(vlim[0])))
    # Check to make sure that the data doesn't contain any values beyond the specified limits, if
    # so, add a greater than or less than sign to the respective label.
    if catdata:
        ticklabels = ["%i" % v for v in np.unique(data)]
    else:
        ticklabels = ticklocs
        # Convert the labels to strings
        ticklabels = [('{:.%ig}' % sigfigs).format(label) for label in ticklabels]
    # Add on the greather than or less than sign if there are values within the dataset that are
    # outside of the provided limits
    if vlim[0] > vmin:
        ticklabels[nticks - 1] = r'$\leq$ %s' % round(vlim[0], sigfigs)
    if vlim[1] < vmax:
        ticklabels[0] = r'$\geq$ %s' % round(vlim[1], sigfigs)

    return vlim, ticklocs, ticklabels


def get_palette(palette, ncat, cmap=True):
    """
    Return a colour palette with the required number of colours. The available palettes are
    ``cat_pastel``, ``cat_vibrant``, and ``cat_dark``. Please refer to the colour palette
    documentation for more information.

    The returned data can be either a list of colours or a Matplotlib ListColormap object.

    Parameters:
        palette (str): Desired colour palette
        ncat (int): The number of categories being plotted
        cmap (bool): Indicate if the retruned pallet is a list of colors or a Matplotlib
            ListColormap

    Returns
    -------
        palette: Matplotlib colormap, list
            if ``cmap`` is ``True`` or a list if ``cmap`` is ``False`` with the required number of
            colours

    Examples:
        To get a list of colour palettes within pygeostat in python, call the following function:

        >>> gs.cmaps.avail_palettes

    .. codeauthor:: pygeostat development team 2015-10-08
    """
    from . cmaps import cat_palettes

    # Check to make sure the number of colours needed doesn't exceed the number the palette selected
    # can provide.
    if palette in ['cat_pastel', 'cat_vibrant'] and ncat > 12:
        print("The colour palette '%s' can only accomodate up to 12 catagories" % palette)
        return
    if palette in ['cat_dark'] and ncat > 6:
        print("The colour palette '%s' can only accomodate up to 6 catagories" % palette)
        return
    # Get the required colour palette
    colours = cat_palettes[palette][:ncat]
    if cmap:
        palette = ListedColormap(colours, name=palette)
    else:
        palette = colours

    return palette


def get_cmap(colormap):
    """
    Return a colormap that is saved within pygeostat. Available colormaps are ``topo1``,
    and ``topo1``. These colormaps are used for hillshading.

    Parameters:
        colormap (str): Desired colormap

    Returns:
        cmap (cmap): Matplotlib colormap

    Examples:
        To get a list of colormaps within pygeostat in python, call the following function:

        >>> gs.get_cmap('topo1')

    .. codeauthor:: pygeostat development team - 2015-10-08
    """
    from .cmaps import cont_cmaps

    # Grab the desired colormap
    cmap = ListedColormap(cont_cmaps[colormap], name=colormap)

    return cmap


def catcmapfromcontinuous(contcmap, ncats, offset=0.5):
    """
    Use the continuous colormap to get a set of categorical values returned in the
    proper format for the other plotting routines in pygeostat from ListedColormap()

    Parameters:
        contcmap (str): valid matplotlib continuous colormap string, like 'jet' or 'spectral'
        ncats (int): number of categories passed to the function
        offset (float): offsets the colors so as they are not the end member colors

    Returns:
        cmap (cmap): Matplotlib colormap

    .. codeauthor:: pygeostat development team 2016-05-18
    """

    from . cmaps import avail_cmaps

    if contcmap in avail_cmaps:
        cm = get_cmap(contcmap)
    else:
        cm = plt.cm.get_cmap(contcmap)
    colors = []
    for num in range(ncats):
        colors.append(cm((num + offset) / (ncats)))

    cmap = ListedColormap(colors, name=contcmap)
    return cmap


def get_label(data):
    """Small utility to grab label information from different types of data sources

    Parameters:
        data: Tidy (long-form) dataframe where each column is a variable and each row is an
            observation. Pandas dataframe or numpy array

    Returns:
        label (str): the name of the column

    .. codeauthor:: pygeostat development team
    """
    label = None
    if hasattr(data, 'columns'):
        if len(data.columns) == 1:
            label = data.columns[0]
    if hasattr(data, 'name'):
        label = data.name

    return label


def clrmplmem():
    """
    Clear the current plot from memory.

    When looping the generation of plots, matplotlib does not remove data from memory until it is
    needed, even if the namespace used is repeated. Instead of clearing the data, the namespace is
    simply remaped. This function uses a series of commands that will dump physical memory.

    Example:

    >>> for var in variables:
    >>>     gs.slice_plot(data[var], griddef=griddef)
    >>>     gs.clrmplmem()

    .. codeauthor:: pygeostat development team - 2016-03-22
    """
    fig = plt.gcf()
    fig.clf()
    plt.close()
    gc.collect()


def addxticks(fig, nrow, ncol, lastax, axextents=None):
    """
    A small utility function called from many of the plotting functions. This adds xticklabels to
    the last subplot in a column when it is not in the last row.

    .. codeauthor:: Ryan Martin and pygeostat development team - 2016-04-18
    """
    import pygeostat as gs

    #  Add xticklabels to the last subplot in a column when it is not in the last row
    renderer = fig.canvas.get_renderer()
    fig.draw(renderer)
    xaxis = fig.axes[ncol * (nrow - 1)].get_xaxis()
    xticklocs = []
    xticklabels = []
    for text in xaxis.get_ticklabels():
        xticklocs.append(text.get_position()[0])
        xticklabels.append(text.get_text())
    xticklocs = np.array(xticklocs)
    if axextents is not None:
        inds = (xticklocs >= axextents[0]) & (xticklocs <= axextents[1])
        xticklabels = np.array(xticklabels)
        xticklocs = xticklocs[inds]
        xticklabels = xticklabels[inds]
    for i in range(lastax, nrow * ncol):  # iax of axes that need xticklabels
        iax = i - ncol
        xaxis = fig.axes[iax].get_xaxis()
        xaxis.tick_bottom()
        xaxis.set_ticks(xticklocs)
        xaxis.set_ticklabels(xticklabels)


def get_supaxislocs(fig, nrow, ncol, figsize, pad):
    """
    A small utility function called from many of the plotting functions. This gets the required
    coordinates for super axis labels.

    Parameters:
        fig (mpl.fig): Matplotlib figure
        nrow (int): Number of rows
        ncol (int): Number of Columns
        figsize (tuple): Figure size (width, height)
        pad (float): Separation between items and the output coordinates

    Returns
    -------
        xmin: float
            Minimum coordinate of figure in x-axis
        xmid: float
            Center coordinate of figure in x-axis
        ymin: float
            Minimum coordinate of figure in y-axis
        ymid: float
            Center coordinate of figure in y-axis
        ymax: float
            Maximum coordinate of figure in y-axis

    .. codeauthor:: pygeostat development team 2016-04-18
    """
    import pygeostat as gs

    renderer = fig.canvas.get_renderer()
    fig.draw(renderer)
    dpi = fig.dpi
    axes = fig.axes
    #  Get the middle coordinate in the x-axis of the grid
    if ncol % 2 == 0:
        #  Get the max x value from the mid left axis
        iax = int(ncol / 2 - 1)
        bbox = axes[iax].get_tightbbox(renderer)
        xmax = bbox.xmax / figsize[0] / dpi
        #  Get the min x value from the mid right axis
        bbox = axes[iax + 1].get_tightbbox(renderer)
        xmin = bbox.xmin / figsize[0] / dpi
        #  Calculate the xmid coordinate
        xmid = ((xmax - xmin) / 2 + xmin)
    else:
        #  Get the middle coordinate from the middle axis
        iax = int((ncol / 2) - 0.5)
        bbox = axes[iax].get_tightbbox(renderer)
        xmin = bbox.xmin / figsize[0] / dpi
        xmax = bbox.xmax / figsize[0] / dpi
        #  Calculate the xmid coordinate
        xmid = ((xmax - xmin) / 2 + xmin)
    #  Get the middle coordinate in the y-axis of the grid
    if nrow % 2 == 0:
        #  Get the max y coord from the middle bottom axis
        iax = int((nrow / 2)) * ncol
        bbox = axes[iax].get_tightbbox(renderer)
        ymax = bbox.ymax / figsize[1] / dpi
        #  Get the min y value from the middle top axis
        iax = (int((nrow / 2)) - 1) * ncol
        bbox = axes[iax].get_tightbbox(renderer)
        ymin = bbox.ymin / figsize[1] / dpi
        #  Calculate the ymid coordinate
        ymid = ((ymax - ymin) / 2 + ymin)
    else:
        #  Get the middle coordinate from the middle axis
        iax = int((nrow / 2) - 0.5) * ncol
        bbox = axes[iax].get_tightbbox(renderer)
        ymin = bbox.ymin / figsize[1] / dpi
        ymax = bbox.ymax / figsize[1] / dpi
        #  Calculate the ymid coordinate
        ymid = ((ymax - ymin) / 2 + ymin)
    #  Get the top of the axes for the suptitle:
    bbox = axes[0].get_tightbbox(renderer)
    ymax = bbox.ymax / figsize[1] / dpi
    ymax = ymax + pad
    #  Get the limit of the xticklabels
    iax = ncol * (nrow - 1)
    bbox = axes[iax].xaxis.get_ticklabel_extents(renderer)[0]
    ymin = (bbox.ymin / figsize[1] / dpi) - pad
    #  Get the x limit of the yticklabels
    bbox = axes[0].yaxis.get_ticklabel_extents(renderer)[0]
    xmin = (bbox.xmin / figsize[0] / dpi) - pad

    return xmin, xmid, ymin, ymid, ymax


def _get_cmap(cmap, catdata, ncat):
    '''Determine the colormap based on the catdata boolean, cmap value and Parameters.
    Used by location_plot'''

    from . cmaps import avail_palettes, avail_cmaps
    from matplotlib.colors import LinearSegmentedColormap
    if catdata:
        if cmap is None:
            cmap = Parameters['plotting.cmap_cat']
            if isinstance(cmap, dict):
                cmap = list(cmap.values())
                cmap = LinearSegmentedColormap.from_list('cmap_cat', cmap, N=len(cmap))
        if not isinstance(cmap, mpl.colors.ListedColormap):
            if cmap in avail_palettes:
                cmap = get_palette(cmap, ncat)
            else:
                cmap = catcmapfromcontinuous(cmap, ncat)
    else:
        if cmap is None:
            cmap = Parameters['plotting.cmap']
        if not isinstance(cmap, mpl.colors.ListedColormap):
            if cmap in avail_cmaps:
                cmap = get_cmap(cmap)
            elif cmap in plt.colormaps():
                cmap = plt.cm.get_cmap(cmap)
            else:
                pass  # this causes MPL to raise an exception and give a list of valid cmaps
    return cmap


def setup_plot(ax, cbar=None, figsize=None, cax=None, aspect=None):
    '''A small utility function called from many of the plotting functions. This will set up a
    matplotlib plot instance based on whether an axis is passed or not.

    Parameters:
        ax (mpl.axis): Matplotlib axis to plot the figure
        cbar (bool): Indicate if a colorbar should be plotted or not
        figsize (tuple): Figure size (width, height)
        cax: Matplotlib.ImageGrid.cbar_axes object
        aspect (bool, str, float): Bool for creating axes, str or float
            for existing axes

    Returns:
        fig (mpl.plt.fig): Matplotlib figure
        ax (mpl.axis): Matplotlib axis to plot the figure
        cax: Matplotlib.ImageGrid.cbar_axes object
    '''
    from mpl_toolkits.axes_grid1 import ImageGrid
    from mpl_toolkits.axes_grid1 import axes_divider

    if ax is None:
        # Setup up a new plot
        fig = plt.figure(figsize=figsize)
        cbar_mode = None
        if cax is None:
            if cbar:
                cbar_mode = 'single'
        if aspect is None:
            aspect = True
        imggrid = ImageGrid(fig, 111, nrows_ncols=(1, 1), axes_pad=0.07,
                            cbar_mode=cbar_mode, cbar_size=0.075, aspect=aspect)
        ax = imggrid[0]
        if cax is None:
            cax = imggrid.cbar_axes[0]
    elif hasattr(ax, "cax"):
        cax = ax.cax
        fig = plt.gcf()
    elif cbar and not isinstance(ax, axes_divider.LocatableAxes) and cax is None:
        fig, cax = get_cbar_axis(ax, cax)
    elif cbar:
        fig = plt.gcf()
        if hasattr(ax, 'cax'):
            cax = ax.cax
        if cax is None:
            raise ValueError("A colorbar axes `cax` must be passed as the passed `ax` cannot be"
                             " divided.")
    else:
        fig = plt.gcf()
    return fig, ax, cax


# ----------------------------------------------------------------------------------------------
# This code was developed by RMS, October 2019, and ported to  pygeostat as maintenance
# support for strange behavior experienced in matplotlib
#
_custom_cbar_width = 0.008
_custom_cbar_pad = 0.008


def _get_mpl_cbar_callback(ax):
    def cax_pos_callback(event):
        ax_rect = ax.get_position()
        ax.cax.set_position([ax_rect.x1 + _custom_cbar_pad,
                             ax_rect.y0, _custom_cbar_width,
                             ax_rect.height])
    return cax_pos_callback


def get_cbar_axis(ax, cax=None):
    """Add a cax (replaces make_axes_locatable)

    related to: https://github.com/matplotlib/matplotlib/issues/4282

    .. codeauthor:: RMS, 2019
    """
    fig = plt.gcf()
    if cax is None:
        cax = getattr(ax, 'cax', None)
    if cax is None:
        ax_rect = ax.get_position()
        cax = fig.add_axes([ax_rect.x1 + _custom_cbar_pad,
                            ax_rect.y0, _custom_cbar_width,
                            ax_rect.height])
        ax.cax = cax
        fig.canvas.mpl_connect('draw_event', _get_mpl_cbar_callback(ax))
    return fig, cax
# end RMS
# ----------------------------------------------------------------------------------------------


def format_subplot_axis(fig, ax, plotformat_dict, cbar=None):
    """
    A small utility function that will allow you to pass a dictionary for formatting such things as
    the fontsize and color of subplot axis

    Parameters:
        fig (matplotlib.fig): the figure for the subplots
        ax (matplotlib.ax): the matplotlib axis for the subplots
        cbar (matplotlib.fig): the cbar figure used in pygeostat plotting functions
        plotformat_dict (dict): dictionary used to format the axis

    Returns:
        fig, ax
        fig, ax, cbar

    Dictionary Keys:

        'fontsize': Dictionary

            * 'title' (int): change title fontsize
            * 'cbar_title' (int): change cbar title fontsize
            * 'xaxis_lable' (int): change xaxis label fontsize
            * 'xaxis_ticklabels' (int): change xaxis tick labels fontsize
            * 'yaxis_label' (int): change yaxis label fontsize
            * 'yaxis_ticklabels' (int): change taxis tick labels fontsize
            * 'cbar_ticklabels' (int): change cbar tick labels fontsize

        'color': Dictionary

            * 'title' (str): change title fontsize
            * 'cbar_title' (str): change cbar title fontsize
            * 'xaxis_lable' (str): change xaxis label fontsize
            * 'xaxis_ticklabels' (str): change xaxis tick labels fontsize
            * 'yaxis_label' (str): change yaxis label fontsize
            * 'yaxis_ticklabels' (str): change taxis tick labels fontsize
            * 'cbar_ticklabels' (str): change cbar tick labels fontsize

    .. codeauthor:: Tyler Acorn - Jan 03, 2018
    """
    if 'fontsize' in plotformat_dict:
        if isinstance(plotformat_dict['fontsize'], dict):
            fs_dict = plotformat_dict['fontsize']
            if 'title' in fs_dict:
                ax.title.set_fontsize(fs_dict['title'])
            if cbar and 'cbar_title' in fs_dict:
                cbar.ax.yaxis.label.set_fontsize(fs_dict['cbar_title'])
            if 'xaxis_label' in fs_dict:
                ax.xaxis.label.set_fontsize(fs_dict['xaxis_label'])
            if 'xaxis_ticklabels' in fs_dict:
                ax.tick_params(axis='x', labelsize=fs_dict['xaxis_ticklabels'])
            if 'yaxis_label' in fs_dict:
                ax.yaxis.label.set_fontsize(fs_dict['yaxis_label'])
            if 'yaxis_ticklabels' in fs_dict:
                ax.tick_params(axis='y', labelsize=fs_dict['yaxis_ticklabels'])
            if cbar and 'cbar_ticklabels' in fs_dict:
                cbar.ax.tick_params(axis='y', labelsize=fs_dict['yaxis_ticklabels'])
    if 'color' in plotformat_dict:
        if isinstance(plotformat_dict['color'], dict):
            c_dict = plotformat_dict['color']
            if 'title' in c_dict:
                ax.title.set_color(c_dict['title'])
            if cbar and 'cbar_title' in fs_dict:
                cbar.ax.yaxis.label.set_color(c_dict['cbar_title'])
            if 'xaxis_label' in c_dict:
                ax.xaxis.label.set_color(c_dict['xaxis_label'])
            if 'xaxis_ticklabels' in c_dict:
                ax.tick_params(axis='x', color=c_dict['xaxis_ticklabels'])
            if 'yaxis_label' in c_dict:
                ax.yaxis.label.set_color(c_dict['yaxis_label'])
            if 'yaxis_ticklabels' in c_dict:
                ax.tick_params(axis='y', color=c_dict['yaxis_ticklabels'])
            if cbar and 'cbar_ticklabels' in c_dict:
                cbar.ax.tick_params(axis='y', color=c_dict['yaxis_ticklabels'])

    if cbar:
        return fig, ax, cbar
    else:
        return fig, ax


def scalebar(x, y, scale, length, ax, img, height=0.005, units='km', pad=0.02):
    """
    A small utility function called from many of the plotting functions. This will Add a scale bar
    to a plot. Requires an Image object such as imshow.

    Parameters:
        x (float): X coordinate of the scale bar in figure space
        y (float): Y coordinate of the scale bar in figure space
        scale (float): Side of each pixel in the units specified by the parameter ``units``
        length (float): Length of the scale bar in the units specified by the parameter ``units``
        ax (mpl.ax): Matplotlib axis to plot the scale bar onto
        img (mpl.AxesImage): Matplotlib AxesImage object. Produced by from imshow()
        height (float): Height of the scale bar as a percentage of plot in y direction in figure
            space
        units (str): Units of the scale bar
        pad (float): Padding between the scale bar and the annotation

    .. codeauthor:: pygeostat development team - 2016-06-13
    """
    # Get some extents
    extents = img.get_extent()
    xsiz = (abs(extents[1]) + abs(extents[0]))
    ysiz = (abs(extents[2]) + abs(extents[3]))
    xmax = max(extents[1], extents[0])
    ymax = max(extents[2], extents[3])
    xmid = xmax - (xsiz * (1 - x))
    pad = ysiz * pad
    height = ysiz * height
    if extents[2] > extents[3]:
        ymid = ymax - (ysiz * y)
        ymin = ymid + (height / 2)
        txty = ymin + pad
    else:
        ymid = ymax - (ysiz * (1 - y))
        ymin = ymid - (height / 2)
        txty = ymin - pad
    length_px = length / scale
    xmin = xmid - (length_px / 2)
    # Make the scale bar rectangle
    rectangle = mpl.patches.Rectangle((xmin, ymin), length_px, height)
    rectangle.set(edgecolor='k', facecolor='k')
    ax.add_patch(rectangle)
    # Add tannotatio
    ax.text(xmid, txty, '%s %s' % (length, units), ha='center', va='top')


def smart_annotate(ax, x, y, labels, k=0.002):
    """
    A small utility function that can be called from a plotting funtion. This annotates a scatter
    plot and moves the annotations around so they do not overlap. Works okay...

    Will need to be made more flexible in the future. Works for gs.mdsplt() at the moment.

    Retrieved and modified from: http://stackoverflow.com/a/34697108

    Parameters:
        ax (mpl.axis): Matplotlib axis object the scatter plot is plotted on
        x: 1-D array with the x-coordinates used for the data in the scatter plot
        y: 1-D array with the y-coordinates used from the data in the scatter plot
        labels (list): Labels to annotate the scatter plot with
        k (float): Scalar for the distance from the data the annotation is placed

    .. codeauthor:: pygeostat development team - 2016-05-30
    """
    import networkx as nx
    G = nx.DiGraph()
    data_nodes = []
    init_pos = {}
    for xi, yi, label in zip(x, y, labels):
        data_str = 'data_{0}'.format(label)
        G.add_node(data_str)
        G.add_node(label)
        G.add_edge(label, data_str)
        data_nodes.append(data_str)
        init_pos[data_str] = (xi, yi)
        init_pos[label] = (xi, yi)
    pos = nx.spring_layout(G, pos=init_pos, fixed=data_nodes, k=k)
    # undo spring_layout's rescaling
    pos_after = np.vstack([pos[d] for d in data_nodes])
    pos_before = np.vstack([init_pos[d] for d in data_nodes])
    scale, shift_x = np.polyfit(pos_after[:, 0], pos_before[:, 0], 1)
    scale, shift_y = np.polyfit(pos_after[:, 1], pos_before[:, 1], 1)
    shift = np.array([shift_x, shift_y])
    for key, val in pos.items():
        pos[key] = (val * scale) + shift

    for label, data_str in G.edges():
        ax.annotate(label,
                    xy=pos[data_str], xycoords='data',
                    xytext=pos[label], textcoords='data',
                    arrowprops=dict(arrowstyle="-",
                                    shrinkA=0, shrinkB=0,
                                    connectionstyle="arc3",
                                    color='k'))
    # expand limits
    all_pos = np.vstack(pos.values())
    x_span, y_span = np.ptp(all_pos, axis=0)
    mins = np.min(all_pos - x_span * 0.02, 0)
    maxs = np.max(all_pos + y_span * 0.02, 0)
    ax.set_xlim([mins[0], maxs[0]])
    ax.set_ylim([mins[1], maxs[1]])


def get_statblk(stat_blk, statsets, statlist, stat_xy):
    """
    A small utility function called from many of the plotting functions. Extracts the required
    statistics and return a string to plot and alignment settings.

    If automatic alignment is not desired, pass the string ``'noalign'`` through ``stat_blk``.

    Note done

    Returns:
        txtstats, stat_xy, ha, va

    .. codeauthor:: pygeostat development team - 2016-07-22
    """
    # Handle input
    if isinstance(stat_blk, str):
        stat_blk = [stat_blk]
    presets = list(statsets.keys())
    stats = list(statlist.keys())
    if 'noalign' in stat_blk:
        align = False
        stat_blk.remove('noalign')
    elif stat_xy in ['bottom', 'top']:
        align = False
    else:
        align = True
    # Sanity checks
    if len(stat_blk) == 0:
        raise ValueError("No parameter was passed to `stat_blk`")
    for stat in stat_blk:
        if stat not in (presets + stats):
            raise ValueError("The following value is not a valid `stat_blk` parameter: %s\n "
                             "Try one of: %s" % (stat, (presets + stats)))
    # Get the right statistics
    if stat_blk[0] in presets:
        stat_blk = [statlist[stat] for stat in statsets[stat_blk[0]]]
    else:
        stat_blk = [statlist[stat] for stat in stat_blk]
    txtstats = '\n'.join(stat_blk)
    # Set-up the alignment
    if align:
        if stat_xy[0] > 0.5:
            ha = 'right'
        else:
            ha = 'left'
        if stat_xy[1] > 0.5:
            va = 'top'
        else:
            va = 'bottom'
    elif stat_xy == 'bottom':
        va = 'bottom'
        ha = 'left'
        stat_xy = (1, 0.05)
    elif stat_xy == 'top':
        va = 'top'
        ha = 'left'
        stat_xy = (1, 0.95)
    else:
        ha = 'left'
        va = 'top'

    return (txtstats, stat_xy, ha, va)


def _set_stat_fontsize(stat_fontsize):
    '''
    Determine the stat_fontsize based on the provided parameter, Parameters
    '''
    if stat_fontsize is None:
        stat_fontsize = Parameters['plotting.stat_fontsize']
    if stat_fontsize is None:
        stat_fontsize = mpl.rcParams['font.size']
    else:
        try:
            if stat_fontsize <= 0:
                raise ValueError('stat_fontsize must be greater than 0!')
            elif stat_fontsize < 1:
                stat_fontsize = mpl.rcParams['font.size'] * stat_fontsize
            # Otherwise, stat_fontsize remains the stat_fontsize
        except:
            raise ValueError('stat_fontsize must be a float or integer!')
    return stat_fontsize


def applytickbins(ax, ntickbins=None):
    """
    Quick function to deal with tuple or integer input

    Parameters:
        ax (mpl.axes): the axis to apply the tickbins too
        ntickbins (int, tuple): either an int (applied to both) or tuple for x and y respectively
    """
    if ntickbins is not None:
        xbins, ybins = None, None
        try:
            iter(ntickbins)
            if len(ntickbins) == 2:
                xbins, ybins = ntickbins
            else:
                xbins, ybins = 5, 5
        except Exception:
            pass
        if isinstance(ntickbins, int):
            xbins, ybins = ntickbins, ntickbins
        if xbins is not None:
            ax.locator_params(axis='x', nbins=xbins)
        if ybins is not None:
            ax.locator_params(axis='y', nbins=ybins)


def format_plot(ax, xlabel=None, ylabel=None, title=None, grid=None, axis_xy=None,
              xlim=None, ylim=None, logx=None, logy=None):
    '''
    Format a plot with common properties such as labels and grid/axis visibility. Not used by the
    spatial plotting function.

    Parameters:
        ax(Matplotlib axis handle): axis handle to modify

    Keyword arguments:
        xlabel(str): label of the x-axis
        ylabel(str): label of the y-axis
        title(str): plot title
        grid(bool): plots the major grid lines if True. Based on Parameters['plotting.grid']
            if None.
        axis_xy(bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
        xlim(tuple): set x-axis limits to xlim[0] and xlim[1]
        ylim(tuple): set y-axis limits to ylim[0] and ylim[1]
        logx, logy (bool): if true, convert each axis to log-scale

    Return:
        ax(Matplotlib axis handle): modified axis handle
    '''
    from matplotlib.ticker import StrMethodFormatter
    # Format the axis labels
    if xlabel is not None:
        if xlabel is not False:
            ax.set_xlabel(xlabel, fontsize=None)
    if ylabel is not None:
        if ylabel is not False:
            ax.set_ylabel(ylabel, fontsize=None)
    if title is not None:
        if title is not False:
            ax.set_title(title, fontsize=None)
    # Format the axis splines
    _format_axis_xy(ax, axis_xy)
    # Format the grid lines
    _format_grid(ax, grid)
    # deal with log axes and ticks
    sigfigs = Parameters.get('plotting.sigfigs')
    if sigfigs is None:
        sigfigs = 3
    fmtstr = '{{x:.{}g}}'.format(sigfigs)
    if logx:
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(StrMethodFormatter(fmtstr))
    if logy:
        ax.set_yscale('log')
        ax.yaxis.set_major_formatter(StrMethodFormatter(fmtstr))
    # Set the x and y limits
    if xlim is not None:
        if logx and xlim[0] <= 0:
            xlower = 1e-6
        else:
            xlower = xlim[0]
        ax.set_xlim(xlower, xlim[1])
    if ylim is not None:
        if logy and ylim[0] <= 0:
            ylower = 1e-6
        else:
            ylower = ylim[0]
        ax.set_ylim(ylower, ylim[1])

    return ax


def _format_axis_xy(ax, axis_xy=None):
    '''
    Private function for the handling axis_xy, which converts the axis to
        GSLIB-style axis visibility (only left and bottom visible) if
        axis_xy is True.

    Parameters:
        ax(matplotlib axis handle): axis to modify
        axis_xy(bool): convert to axis_xy if True

    Return:
        ax(Matplotlib axis handle): modified axis handle
    '''
    if axis_xy is None:
        axis_xy = Parameters['plotting.axis_xy']
    if axis_xy:
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)


def _format_grid(ax, grid=None, below=False):
    '''
    Private function for the handling of grid lines. Matplotlib kwargs
        could be added in the future for permitting custom grid lines,
        as was done in the previous pygeostat version (commented out below)

    Parameters:
        ax(matplotlib axis handle): axis to append grid lines to (or not)
        grid(bool): add grid lines, or not...
        below(bool): place the grid lines below (Ture) or above (False)
            the plot content

    Return:
        ax(Matplotlib axis handle): modified axis handle
    '''
    if grid is None:
        grid = Parameters['plotting.grid']
    if grid:
        # The grid lines won't be seen in many plots if not above...
        # Could consider having this as an option
        ax.set_axisbelow(below)
        # The linestyle, width and color are available in rcParams and should, perhaps,
        # not be mangled with. That said, the pygeostat setting below looks way better than
        # the matplotlib default so leaving for now
        ax.grid(True, which='major', color='k', linestyle=':', lw=0.2)


def _spatial_labels(ax, orient, griddef=None, slice_number=0, title=None, xlabel=None, ylabel=None,
                    unit=None, sigfigs=3):
    '''
    A small utility function called from plotting functions that involve maps and cross-sections.
    It provides the appropriate labels and title according to a variety of parameter options.

    Parameters:
        ax (mpl.axis): Matplotlib axis to plot the figure
        orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted
            values
        griddef (GridDef): A pygeostat GridDef class created using
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`
        slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
            plot
        title (str): Title for the plot. If left to it's default value of ``None`` or is set to
            ``True``, a logical default title will be generated for 3-D data. Set to ``False`` if
            no title is desired.
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        unit (str): Unit to place inside the axis label parentheses

    Returns:
        ax (mpl.axis): Matplotlib axis to plot the figure
    '''
    from ..data.grid_definition import GridDef
    if unit is None:
        unit = Parameters['plotting.unit']
    if unit is None or unit == '':
        unit = ''
    else:
        unit = ' ({})'.format(unit)
    if xlabel:
        ax.set_xlabel(xlabel, fontsize=None)
    elif xlabel is None:
        if orient == 'xy':
            ax.set_xlabel(Parameters['plotting.xname'] + unit, fontsize=None)
        elif orient == 'xz':
            ax.set_xlabel(Parameters['plotting.xname'] + unit, fontsize=None)
        elif orient == 'yz':
            ax.set_xlabel(Parameters['plotting.yname'] + unit, fontsize=None)
    if ylabel:
        ax.set_ylabel(ylabel)
    elif ylabel is None:
        if orient == 'xy':
            ax.set_ylabel(Parameters['plotting.yname'] + unit, fontsize=None)
        elif orient == 'xz':
            ax.set_ylabel(Parameters['plotting.zname'] + unit, fontsize=None)
        elif orient == 'yz':
            ax.set_ylabel(Parameters['plotting.zname'] + unit, fontsize=None)
    if unit != '':
        unit = unit[2:-1]
    if isinstance(title, str):
        ax.set_title(title, fontsize=None)
    elif (title is None or title is True) and isinstance(griddef, GridDef):
        crossd = list(set(['x', 'y', 'z']) - set(list(orient)))[0]
        d3 = False
        if crossd == 'x' and griddef.nx > 1:
            d3 = True
        if crossd == 'y' and griddef.ny > 1:
            d3 = True
        if crossd == 'z' and griddef.nz > 1:
            d3 = True
        if d3:
            if orient in ['xz', 'yz']:
                cross = griddef.get_slice_coordinate(orient, slice_number)
                if crossd is 'y':
                    crossd = Parameters['plotting.yabbrev']
                else:
                    crossd = Parameters['plotting.xabbrev']
                title = 'Cross Section: %s %s %s' % (round(cross, sigfigs), unit, crossd)
            else:
                plan = griddef.get_slice_coordinate(orient, slice_number)
                title = 'Plan View: %s %s' % (round(plan, sigfigs), unit)
        else:
            title = None
    else:
        title = None
    if title:
        ax.set_title(title, fontsize=None)
    return ax


def _spatial_pointdata(data, orient, x=None, y=None, z=None, griddef=None):
    '''Used by slice_plot and location_plot'''
    import pandas as pd
    from ..data.data import DataFile
    from ..data.grid_definition import GridDef
    import pygeostat as gs
    if not any([isinstance(data, pd.DataFrame), isinstance(data, DataFile)]):
        raise ValueError("The parameter data must be either a gs.DataFile or a pd.DataFrame")
    if isinstance(data, DataFile):
        if not griddef:
            if isinstance(data.griddef, GridDef):
                griddef = data.griddef
        if x is None:
            x = data.x
        if y is None:
            y = data.y
        if z is None:
            z = data.z
        data = data.data
    if orient == 'xy':
        if any([x is None, y is None]):
            raise ValueError("`orient` is set to `'xy'` yet the column ID for `x` and `y` has"
                             " not been set and could not be automatically retrieved.")
    elif orient == 'xz':
        if any([x is None, z is None]):
            raise ValueError("`orient` is set to `'xz'` yet the column ID for `x` and `z` has"
                             " not been set and could not be automatically retrieved.")
    elif orient == 'yz':
        if any([y is None, z is None]):
            raise ValueError("`orient` is set to `'yz'` yet the column ID for `y` and `z` has"
                             " not been set and could not be automatically retrieved.")
    else:
        raise ValueError("An `orient` setting of %s is not permissible" % orient)
    return data, x, y, z, griddef


def _spatial_griddata(data, var, griddef):
    '''
    Private function that parses the data, var and griddef, returning the data
    and griddef to apply. Used by slice_plot and contourplt.
    '''
    from .. data.data import DataFile

    # Handle DataFile input
    if isinstance(data, DataFile):
        if griddef is None:
            griddef = data.griddef
        data = data.data
    # Ensure that we have a griddef
    if griddef is None:
        raise Exception('Error: griddef must be provided if data is not a DataFile object\n'
                        '       with the associated grid definition (data.griddef)')
    # Handle panda input
    if hasattr(data, 'values'):
        if isinstance(var, str) and data.shape[1] > 1:
            data = data[var].values
        else:
            data = data.values
            if data.ndim > 1:
                if isinstance(var, int):
                    # If var is provided as a scalar, then assume it's the column number
                    data = data[:, var]
                else:
                    data = data[:, 0]
    if data.ndim > 1:
        if data.shape[1] > 1:
            raise ValueError('could not coerce data into a 1-D array!')
    return data, griddef


def _spatial_slice(data, var, x, y, z, griddef, orient, slice_number, slicetol, dhid=None):
    '''Used by slice_plot and location_plot'''
    from .. datautils.utils import slicescatter
    from .. data.grid_definition import GridDef
    if not isinstance(griddef, GridDef) and slicetol is not None:
        raise ValueError("`griddef` must be set if using `slicetol`")
    # If a slice_number is passed but no slice tolerence. set slicetol to griddef cell size
    try:
        if slice_number >= 0:
            if slicetol is None:
                if orient == 'xy':
                    if griddef.nz > 1:
                        slicetol = griddef.zsiz / 2
                    else:
                        slicetol = None
                if orient == 'xz':
                    if griddef.ny > 1:
                        slicetol = griddef.ysiz / 2
                    else:
                        slicetol = None
                    slicetol = griddef.ysiz / 2
                if orient == 'yz':
                    if griddef.nx > 1:
                        slicetol = griddef.xsiz / 2
                    else:
                        slicetol = None
    except:
        pass
    # if no slice_number is passed plot everything. i.e. slice_number = 1, slicetol = None
    if slice_number is None:
        slice_number = 1
        slicetol = None
    # If slicetol is negative set it to plot everything
    try:
        if slicetol < 0:
            slicetol = None
    except:
        pass
    # Only slice if needed
    if slicetol is None:
        pointview = data
    else:
        pointview = slicescatter(data, orient, slice_number, slicetol, griddef, x, y, z)
        if len(pointview) == 0:
            #raise ValueError("No data was found within the specified slice tolerance")
            if dhid is None:
                return None, None, None
            else:
                return None, None, None, None
                
    figx, figy, figz = _spatial_orient2fig(orient, x, y, z)
    if figz is not None:
        pointsort = np.argsort(pointview[figz].values)
        pointx = pointview[figx].values[pointsort]
        pointy = pointview[figy].values[pointsort]
        try:
            pointvar = pointview[var].values[pointsort]
        except:
            pointvar = None
        try:
            point_dh = pointview[dhid].values[pointsort]
        except:
            point_dh = None

    else:
        pointx = pointview[figx]
        pointy = pointview[figy]
        try:
            pointvar = pointview[var]
        except:
            pointvar = None
        
        try:
            point_dh = pointview[dhid]
        except:
            point_dh = None

    if dhid is None:
        return pointx, pointy, pointvar
    else:
        return pointx, pointy, pointvar, point_dh


def _spatial_orient2fig(orient, x, y, z):
    '''
    Used by a variety of functions. Given orient and the x, y, z cols,
    return figx and figy, which is used for indexing.
    '''
    if orient == 'xy':
        figx = x
        figy = y
        figz = z
    elif orient == 'xz':
        figx = x
        figy = z
        figz = y
    elif orient == 'yz':
        figx = y
        figy = z
        figz = x
    else:
        raise ValueError('orient must be xy, xz or yz!')
    return figx, figy, figz


def _spatial_aspect(xlim, ylim):
    '''
    Simple utility for determining what aspect should be used for spatial plots.
    '''
    xrang = xlim[1] - xlim[0]
    yrang = ylim[1] - ylim[0]
    aspect = 'equal'
    if xrang / yrang > 5:
        aspect = 'auto'
    elif yrang / xrang > 5:
        aspect = 'auto'
    return aspect


def _format_tick_labels(ax, rotateticks=None, nticks=None):
    '''
    A small utility function called from many of the plotting functions. This will format the
    tick labels and rotate them if needed.

    Note - currently mangles padding and alignment according to the original pygeostat settings.

    Parameters:
        ax (mpl.axis): Matplotlib axis to plot the figure
        rotateticks (bool or float tuple): Bool or float values to control axis label rotations.
            See above for more info (x, y).

    '''
    if rotateticks is None:
        rotateticks = Parameters['plotting.rotateticks']
    if rotateticks is None:
        rotateticks = (None, None)
    if nticks is None:
        nticks = Parameters['plotting.nticks']
    try:
        applytickbins(ax, nticks)
    except Exception:
        pass
    ax.tick_params(axis='x', pad=2)
    ax.tick_params(axis='y', pad=1)
    if rotateticks is None:
        # The plots tick labels will not be properly accessible until the figure is "drawn", once
        # the command below is run, ax.get_ticklabel() will actually work properly.
        plt.draw()
        rotateticks = _tickoverlap(ax)
    if isinstance(rotateticks, tuple):
        rotateticks = list(rotateticks)
    # Rotate if required
    rotatex = False
    if isinstance(rotateticks[0], bool) and rotateticks[0]:
        rotateticks[0] = 45
        rotatex = True
    elif rotateticks[0] is not None:
        rotatex = True
    if rotatex:
        xlabels = ax.get_xticklabels()
        for xlabel in xlabels:
            xlabel.set_ha('center')
            xlabel.set_va('top')
            xlabel.set_rotation(rotateticks[0])
    rotatey = False
    if isinstance(rotateticks[1], bool) and rotateticks[1]:
        rotateticks[1] = -45
        rotatey = True
    elif rotateticks[1] is not None:
        rotatey = True
    if rotatey:
        if abs(rotateticks[1]) == 90:
            ax.tick_params(axis='y', pad=0)
        ylabels = ax.get_yticklabels()
        for ylabel in ylabels:
            ylabel.set_ha('right')
            ylabel.set_va('center')
            ylabel.set_rotation(rotateticks[1])
    return ax


def _tickoverlap(ax):
    """
    A small utility function called from many of the plotting functions. This returns booleans
    for whether the x and y tick labels on an Axes overlap as a tuple.

    Parameters:
        ax (ax): A matplotlib axis object with axis tick labels.

    Returns:
        overlaps (tuple bool): Indicators for both axes if the tick labels need to be rotated.

    .. codeauthor:: pygeostat development team - 2015-09-30
    """
    # Set-up a function that completes the check
    def tickcheck(labels):
        """Check to see if labels overlap, returns a boolean value"""
        fig = plt.gcf()
        renderer = fig.canvas.get_renderer()
        fig.draw(renderer)
        if not labels:
            return False
        else:
            try:
                boxes = [label.get_window_extent() for label in labels]
                overlaps = [box.count_overlaps(boxes) for box in boxes]
                return max(overlaps) > 1
            except:
                print("There was an error checking for overlapping axis tick labels. You'll have",
                      "to set the argument manually to rotate them if necessary")
    overlaps = [tickcheck(ax.get_xticklabels()), tickcheck(ax.get_yticklabels())]

    return overlaps


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''desurvey.py: Contains utilities for desurveying drill hole data'''
from __future__ import absolute_import, division, print_function

__author__ = 'pygeostat development team'
__date__ = '2014-04-03'
__version__ = '3.000'

from math import sin, cos, radians, ceil
import numpy as np
import pandas as pd
from ..data import DataFile
import warnings
from collections import OrderedDict
import xml.etree.cElementTree as ET
import copy
from .. pygeostat_parameters import Parameters


class Drillhole(object):
    '''This class contains specific drill hole data and metadata.

    Drill hole classes may be created or generated using pygeostat functions.
    This is primarily used for desurveying drill hole data.

    Parameters:
        holeid (str): Hole Id
        collarx (numeric): x coordinate
        collary (numeric): y coordinate
        collarz (numeric): z coordinate

    .. codeauthor: pygeostat development team 2014-04-03'''

    def __init__(self, holeid=None, collarx=None, collary=None, collarz=None):
        self.holeid = holeid
        self.collarx = float(collarx)
        self.collary = float(collary)
        self.collarz = float(collarz)
        self.data = pd.DataFrame()

    def __str__(self):
        return('Drillhole {} collared at ({},{},{})'.format(self.holeid, self.collarx,
                                                            self.collary, self.collarz))

    def delx(self, dist, inclination, bearing):
        '''Calculate change in X

        Parameters:
            self
            dist (float): distance along direction
            inclination (float)
            bearing (float)

        Returns:
            Value (float): A float value of the change in distance

        .. codeauthor: pygeostat development team 2014-04-03'''
        val = dist * cos(radians(inclination)) * sin(radians(bearing))
        return val

    def dely(self, dist, inclination, bearing):
        '''Calculate change in y

        Parameters:
            self:
            dist (float): distance along direction
            inclination (float)
            bearing (float)

        Returns:
            Value (float): A float value of the change in distance

        .. codeauthor: pygeostat development team 2014-04-03'''
        val = dist * cos(radians(inclination)) * cos(radians(bearing))
        return val

    def delz(self, dist, inclination, bearing):
        '''Calculate change in z

        Parameters:
            self
            dist (float): distance along direction
            inclination (float)
            bearing (float): Not currently used in the calculation

        Returns:
            Value (float): A float value of the change in distance

        .. codeauthor: pygeostat development team 2014-04-03'''
        val = dist * sin(radians(inclination))
        return val

    def getxyz(self, along, null):
        '''Return X, Y, Z values given the length along the drill hole

        Parameters:
            self:
            along (numeric): Distance along the drill hole

        Returns:
            x, y, z (float)

        .. codeauthor: pygeostat development team 2014-04-03'''
        x, y, z = (self.collarx, self.collary, self.collarz)
        # Check some base cases
        if (along < 0.) or (along > float(self.data['Along'][-1:])):
            x, y, z = (null, null, null)
            print('ERROR: This along is invalid ', along, 'for', self.holeid)
        elif along == 0.:
            pass
        else:
            loc = 0.
            idx = 0
            while loc < along:
                try:
                    alongcheck = self.data['Along'][idx + 1]
                except KeyError:
                    alongcheck = 1e21
                    print('WARNING: Hole {} only has 1 survey station, duplicating it'.format(self.holeid))
                    idx = idx - 1
                if along >= alongcheck:
                    dist = self.data['Along'][idx + 1] - loc
                    loc = self.data['Along'][idx + 1]
                    idx += 1
                    x += self.delx(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
                    y += self.dely(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
                    z += self.delz(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
                else:
                    dist = along - loc
                    loc = along
                    idx += 1
                    x += self.delx(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
                    y += self.dely(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
                    z += self.delz(dist, self.data['Inclination'][idx],
                                   self.data['Azimuth'][idx])
        return(x, y, z)


def set_desurvey(collarfl, surveyfl, along_name, azimuth_name, inclination_name):
    '''Sets up for desurveying returning a dictionary of DrillHole objects

    Parameters:
        collarfl (DataFile): pygeostat DataFile with dh, x, y, z all set
        surveyfl (DataFile): pygeostat DataFile with dh set
        along_name (str): variable name of "along" in survey file
        azimuth_name (str): variable name of "azimuth/bearing" in survey file
        inclination_name (str): variable name of "inclination" in survey file

    Returns:
        drillholes (dict): Returns a dictionary of DrillHole objects

    .. codeauthor:: pygeostat development team - 2014-04-03'''
    print(("\nWARNING: drillhole routines are untested with recent Pygeostat changes!\n"
           "           Proceed with caution!"))
    # Check for missing components
    for item in [collarfl.dh, collarfl.x, collarfl.y, collarfl.z,
                 surveyfl.dh, surveyfl.data[along_name],
                 surveyfl.data[azimuth_name], surveyfl.data[inclination_name]]:
        pass
    # Create drill hole objects with collar information
    drillholes = {}
    for rowidx, row in collarfl.data.iterrows():
        drillholes[row[collarfl.dh]] = Drillhole(holeid=row[collarfl.dh], collarx=row[collarfl.x],
                                                 collary=row[collarfl.y], collarz=row[collarfl.z])
    # Get along, azimuth and inclination data for this drill hole
    for dhid in drillholes:
        # Find rows with this drill hole ID
        rows = surveyfl.data[surveyfl.data[surveyfl.dh] == dhid].index.values
        # Check that rows were actually found
        if len(rows) == 0:
            raise Exception('No survey information found for' + drillholes[dhid])
        # Attach the data to the Drillhole
        drillholes[dhid].data = pd.DataFrame({'Along': surveyfl.data[along_name].ix[rows],
                                              'Azimuth': surveyfl.data[azimuth_name].ix[rows],
                                              'Inclination': surveyfl.data[inclination_name].ix[rows]})
        # Sort by along to ensure ordered values and reset the index
        drillholes[dhid].data = drillholes[dhid].data.sort_values(by='Along')\
                                                .reset_index()[['Along', 'Azimuth', 'Inclination']]
    return drillholes


def get_desurvey(datafl, drillholes, inplace=True, x='X', y='Y', z='Z', null=None):
    '''Gets the desurvey of a DataFile given froms and tos.

    Parameters:
        datafl (DataFile): pygeostat DataFile to be desurveyed with one of the options:

                        1) just ifrom or just ito set to desurvey at exactly that distance
                        2) both ifrom and ito set to desurvey halfway in between (ie: midpoint)

        drillholes (dict): dictionary of Drillhole objects, obtained from set_desurvey

    Keyword Args:
        inplace (bool): modifies the datafl.data to have X, Y and Z values
            otherwise returns a pandas dataframe with X, Y and Z

    .. codeauthor: pygeostat development team 2014-04-03'''

    # Create a temporary data file - this allows variable creation
    # without leaving behind a mess or overwriting anything accidentally!
    # There are probably cleaner ways to do this, but desurveying is normally a 1-off.
    print(("\nWARNING: drillhole routines are untested with recent Pygeostat changes!\n"
           "           Proceed with caution!"))
    if null is None:
        null = Parameters['data.null']
    tempfl = pd.DataFrame(datafl.data[datafl.dh])
    # Get the midpoint, or 'along' value to use
    if datafl.ifrom is not None and datafl.ito is not None:
        tempfl['mid'] = (datafl.data[datafl.ito] + datafl.data[datafl.ifrom]) * 0.5
    elif datafl.ifrom is not None and datafl.ito is None:
        tempfl['mid'] = datafl.data[datafl.ifrom]
    elif datafl.ifrom is None and datafl.ito is not None:
        tempfl['mid'] = datafl.data[datafl.ito]
    else:
        raise Exception('Both ifrom and ito are missing from datafl!')

    # Define a local desurvey function
    def desurvey(row):
        'Define a local desurvey function'
        # Get the corresponding drill hole
        drillhole = drillholes[row[datafl.dh]]
        # Get the X,Y,Z coordinates
        xloc, yloc, zloc = drillhole.getxyz(row['mid'], null)
        return pd.Series({x: xloc, y: yloc, z: zloc})
    # Desurvey
    xyzpoints = tempfl.apply(desurvey, axis=1)
    if not inplace:
        return xyzpoints
    else:
        # Assign data
        datafl.data[x] = xyzpoints[x]
        datafl.data[y] = xyzpoints[y]
        datafl.data[z] = xyzpoints[z]
        # Set datafl.x, datafl.y, datafl.z
        if datafl.x is None:
            datafl.x = x
        if datafl.y is None:
            datafl.y = y
        if datafl.z is None:
            datafl.z = z
        return None


def set_comps(datafl, complength):
    '''Returns pandas dataframe with drillhole, compfrom, compto

    Parameters:
        datafl (DataFile): pygeostat DataFile with dh, ifrom and ito set
        complength (numeric): length of composites

    Returns:
        comps (DataFrame): pandas DataFrame with dh, ifrom and ito set for each composite

    .. codeauthor: pygeostat development team 2014-04-03'''
    print(("\nWARNING: drillhole routines are untested with recent Pygeostat changes!\n"
           "           Proceed with caution!"))
    # Get list of drill holes
    dhids = datafl.unique_cats(datafl.data[datafl.dh])
    # Determine start (smallest 'from' value) and end (largest 'to' value) of each dh
    starts = {}
    ends = {}
    nsamples = {}
    totalsamples = 0
    for dhid in dhids:
        starts[dhid] = min(datafl.data[datafl.data[datafl.dh] == dhid][datafl.ifrom])
        ends[dhid] = max(datafl.data[datafl.data[datafl.dh] == dhid][datafl.ito])
        # Determine number of composites in each
        nsamples[dhid] = ceil((ends[dhid] - starts[dhid]) / complength)
        ends[dhid] = nsamples[dhid] * complength + starts[dhid]
        totalsamples += nsamples[dhid]
    # Initialize pandas dataframe
    datadict = OrderedDict()
    datadict[datafl.dh] = [dhid for dhid in range(int(totalsamples))]
    datadict[datafl.ifrom] = np.ones(int(totalsamples))
    datadict[datafl.ito] = np.ones(int(totalsamples))
    comps = pd.DataFrame(datadict)
    # Composites
    startidx = 0
    for dhid in dhids:
        # Pandas slices using INCLUSIVE INDEXING therefore subtract 1 and add 1 later
        endidx = startidx + nsamples[dhid] - 1
        # Set drill hole number, from and to
        comps.loc[startidx:endidx, datafl.dh] = dhid
        comps.loc[startidx:endidx, datafl.ifrom] = np.linspace(starts[dhid],
                                                               ends[dhid] - complength,
                                                               num=nsamples[dhid], endpoint=True)
        comps.loc[startidx:endidx, datafl.ito] = np.linspace(starts[dhid] + complength, ends[dhid],
                                                             num=nsamples[dhid], endpoint=True)
        startidx = endidx + 1
    return comps


def get_comps(comps, datafl, vartypes='continuous', null=None, nprocess=None):
    '''Returns a pandas DataFrame with upscaled composites parallelizing across
       drillholes.

    Parameters:
        datafl (DataFile): pygeostat DataFile with dh, ifrom, ito and at least 1 variable to upscale
        comps (DataFrame): pandas DataFrame with datafl.dh, datafl.ifrom, datafl.ito
        corresponding to the composite locations

    Keyword Args:
        vartypes (str OR dict): 'continuous', 'categorical' or a dictionary of variables like:
        {'Cu':'continuous','Facies','categorial'}
        null: the null value (values less than or equal to this will not be used)

    Returns:
        upscaled (DataFrame): Pandas DataFrame with values of upscaled variable

    .. codeauthor: pygeostat development team 2014-04-03'''
    import multiprocessing as mp
    print(("\nWARNING: drillhole routines are untested with recent Pygeostat changes!\n"
           "           Proceed with caution!"))
    if null is None:
        null = Parameters['data.null']
    # Get a list of drill holes for composites
    dhids = sorted(list(set(comps[datafl.dh])))
    # Determine the number of composites
    ncomps = len(comps[datafl.dh])
    # Create a dictionary of variable upscaling methods if not supplied
    if isinstance(vartypes, str):
        vartypedict = {}
        for var in datafl.data.columns:
            if var not in [datafl.dh, datafl.ifrom, datafl.ito]:
                vartypedict[var] = vartypes.lower()
    else:
        vartypedict = vartypes
    # Determine number of variables
    nvar = len(vartypedict)
    # Initialize a pandas DataFrame for upscaled values
    upscaled = pd.DataFrame(np.ones([ncomps, nvar]) * null, columns=list(vartypedict))
    upscaled[datafl.dh] = comps[datafl.dh]
    upscaled[datafl.ifrom] = comps[datafl.ifrom]
    upscaled[datafl.ito] = comps[datafl.ito]

    # Group data by drill hole ID
    grouped_comps = comps.groupby([datafl.dh], squeeze=True, as_index=False)
    grouped_data = datafl.data.groupby([datafl.dh], squeeze=True, as_index=False)

    # Upscale all drill holes in parallel asynchronously
    if nprocess is None:
        nprocess = Parameters['config.nprocess']
    # Create a pool for processing, calling with None to use all CPUs
    for var, vartype in vartypedict.items():
        # print('Assembling pool for {}...'.format(var))
        pool = mp.Pool(processes=nprocess)
        results = {}
        number_of_valid_dhs = 0
        for dhid, compdf in grouped_comps:
            # Get the data which corresponds to this composite
            try:
                datadf = grouped_data.get_group(dhid)
                number_of_valid_dhs += 1
            except KeyError:
                # No data for this composite drill hole so skip it
                continue
            # Add on the GSLIB call to execute asynchronously
            results[dhid] = pool.apply_async(upscale_worker, (compdf, datadf, var, vartype,
                                                              datafl.ifrom, datafl.ito, null))
        # Close the list of processes for the pool
        pool.close()
        # "Join" the list of processes to execute and wait for completion
        pool.join()
        # Tabulate the results that are not null (ie: a float)
        try:
            all_results = (pd.concat([group.get() for name, group in results.items()
                                      if type(group.get()) is not float]))
        except ValueError:
            # No data were found
            warnings.warn('No valid samples found for variable {}!!!'.format(var))
            return(upscaled)
        # Join this to our database
        upscaled[var] = upscaled[var].replace({null: np.nan})
        upscaled[var].fillna(all_results, inplace=True)
        upscaled[var].fillna(value=null, inplace=True)

    return(upscaled)


def upscale_worker(compdata, dhdata, var, vartype, ifrom, ito, null=None):
    '''Returns the upscaled value contained in [compfrom,compto]

    Parameters:
        compdata (DataFrame): pandas DataFrame with columns for ifrom, ito for the DH of interest
        dhdata (DataFrame): pandas DataFrame with columns for ifrom, ito and var for the DH
            of interest
        var (str): variable name
        vartype (str): 'continuous' or 'categorical'

    Keyword Args:
        null: value to return if no data in interval

    Returns:
        compdata[var]: pandas series of upscaled variable using either:
            continuous - length weighted linear average or
            categorical - most frequent category

    .. codeauthor: pygeostat development team 2014-04-03'''
    if null is None:
        null = Parameters['data.null']
    # Initialize the upscaled variable
    compdata[var] = np.ones(len(compdata[ifrom])) * null
    # Iterate over composite rows
    for compidx, comprow in compdata.iterrows():
        compfrom = comprow[ifrom]
        compto = comprow[ito]
        # Define local function to get length in interval

        def length_in_interval(datarow):
            'finds the length in the interval along the drill hole - tda'
            # Is the variable value valid?
            if datarow[var] <= null:
                return 0.0
            # Case 1 - entire sample interval in composite
            if (datarow[ifrom] >= compfrom) and (datarow[ito] <= compto):
                return datarow[ito] - datarow[ifrom]
            # Case 2 - partially bottom of interval
            elif ((datarow[ifrom] >= compfrom) and (datarow[ifrom] < compto) and
                  (datarow[ito] > compto)):

                return compto - datarow[ifrom]
            # Case 3 - partially in top of interval
            elif ((datarow[ifrom] < compfrom) and (datarow[ito] > compfrom) and
                  (datarow[ito] <= compto)):

                return datarow[ito] - compfrom
            # Case 4 - composite bracketed by sample interval
            elif (datarow[ifrom] <= compfrom) and (datarow[ito] >= compto):
                return compto - compfrom
            # Case 5 - no length in interval
            else:
                return 0.0
        # Get length of each sample in composite
        try:
            dhdata['Length'] = dhdata.apply(length_in_interval, axis=1)
        except ValueError as e:  # No valid data in drill hole found - null assignment!
            return null
        # If no data was found, return null value
        if sum(dhdata['Length']) <= 0.0:
            upscaled = null
        else:
            # Upscale the data
            # - Continuous
            if vartype.startswith('con'):
                upscaled = np.dot(dhdata[var], dhdata['Length']) / sum(dhdata['Length'])
            # - Categorical
            elif vartype.startswith('cat'):
                unique_cats = list(set(dhdata[dhdata['Length'] > 0.0][var]))
                unique_cat_wts = [sum(dhdata[dhdata[var] == cat]['Length']) for cat in unique_cats]
                upscaled = unique_cats[unique_cat_wts.index(max(unique_cat_wts))]
            # - Invalid variable type
            else:
                raise Exception('Invalid vartype', vartype)
        # Save this value
        compdata.loc[compidx, var] = upscaled

    # Return all values for this drill hole
    return(compdata[var])


def serial_upscale(compdh, compfrom, compto, datafl, var, vartype, null=None):
    '''Returns the upscaled value contained in [compfrom,compto]

    Parameters:
        compdh: composite drill hole
        compfrom: composite 'from'
        compto: composite 'to'
        datafl (DataFile): pygeostat DataFile with dh, ifrom, ito and var
        var (str): variable name
        vartype (str): 'continuous' or 'categorical'

    Keyword Args:
        null: value to return if no data in interval

    Returns:
        upscaled: value of upscaled variable using either:
            continuous - length weighted linear average or
            categorical - most frequent category

    .. codeauthor: pygeostat development team 2014-04-03'''
    if null is None:
        null = Parameters['data.null']
    # Get the variable, froms and tos which are in that drill hole
    dhdata = datafl.data[datafl.data[datafl.dh] == compdh][[datafl.ifrom, datafl.ito, var]]
    # Define local function to get length in interval

    def length_in_interval(datarow):
        'finds the length in the interval along the drill hole - tda'
        # Is the variable value valid?
        if datarow[var] <= null:
            return 0.0
        # Case 1 - entire sample interval in composite
        if (datarow[datafl.ifrom] >= compfrom) and (datarow[datafl.ito] <= compto):
            return datarow[datafl.ito] - datarow[datafl.ifrom]
        # Case 2 - partially bottom of interval
        elif ((datarow[datafl.ifrom] >= compfrom) and (datarow[datafl.ifrom] < compto) and
              (datarow[datafl.ito] > compto)):

            return compto - datarow[datafl.ifrom]
        # Case 3 - partially in top of interval
        elif ((datarow[datafl.ifrom] < compfrom) and (datarow[datafl.ito] > compfrom) and
              (datarow[datafl.ito] <= compto)):

            return datarow[datafl.ito] - compfrom
        # Case 4 - composite bracketed by sample interval
        elif (datarow[datafl.ifrom] <= compfrom) and (datarow[datafl.ito] >= compto):
            return compto - compfrom
        # Case 5 - no length in interval
        else:
            return 0.0
    # Get length of each sample in composite
    try:
        dhdata['Length'] = dhdata.apply(length_in_interval, axis=1)
    except ValueError:  # No valid data in drill hole found - null assignment!
        return null
    # If no data was found, return null value
    if sum(dhdata['Length']) <= 0.0:
        return null
    # Upscale the data
    # - Continuous
    if vartype.startswith('con'):
        upscaled = np.dot(dhdata[var], dhdata['Length']) / sum(dhdata['Length'])
    # - Categorical
    elif vartype.startswith('cat'):
        unique_cats = list(set(dhdata[dhdata['Length'] > 0.0][var]))
        unique_cat_wts = [sum(dhdata[dhdata[var] == cat]['Length']) for cat in unique_cats]
        upscaled = unique_cats[unique_cat_wts.index(max(unique_cat_wts))]
    # - Invalid variable type
    else:
        raise Exception('Invalid vartype', vartype)
    return upscaled


def serial_get_comps(comps, datafl, vartypes='continuous', null=None):
    '''Returns a pandas DataFrame with upscaled composites

    Parameters:
        datafl (DataFile): pygeostat DataFile with dh, ifrom, ito and at least 1 variable to upscale
        comps (DataFrame): pandas DataFrame with datafl.dh, datafl.ifrom, datafl.ito
        corresponding to the composite locations

    Keyword Args:
        vartypes (str OR dict): 'continuous', 'categorical' or a dictionary of variables like:
        {'Cu':'continuous','Facies','categorial'}
        null: the null value (values less than or equal to this will not be used)

    Returns:
        upscaled (DataFrame): Pandas DataFrame with values of upscaled variable

    .. codeauthor: pygeostat development team 2014-04-03'''
    if null is None:
        null = Parameters['data.null']
    # Get a list of drill holes for composites
    dhids = sorted(list(set(comps[datafl.dh])))
    # Determine the number of composites
    ncomps = len(comps[datafl.dh])
    # Create a dictionary of variable upscaling methods if not supplied
    if isinstance(vartypes, str):
        vartypedict = {}
        for var in datafl.data.columns:
            if var not in [datafl.dh, datafl.ifrom, datafl.ito]:
                vartypedict[var] = vartypes.lower()
    else:
        vartypedict = vartypes
    # Determine number of variables
    nvar = len(vartypedict)
    # Initialize a pandas DataFrame for upscaled values
    upscaled = pd.DataFrame(np.ones([ncomps, nvar]) * null, columns=list(vartypedict))
    upscaled[datafl.dh] = comps[datafl.dh]
    upscaled[datafl.ifrom] = comps[datafl.ifrom]
    upscaled[datafl.ito] = comps[datafl.ito]
    # Upscale and return
    for var, vartype in vartypedict.items():
        # Define the variable specific upscaling function to apply
        def varupscale(compseries):
            'Define the variable specific upscaling function to apply'
            return(upscale(compseries[datafl.dh], compseries[datafl.ifrom],
                           compseries[datafl.ito], datafl, var, vartype, null))
        # Apply this to the variable
        upscaled[var] = comps.apply(varupscale, axis=1)
    return upscaled


def fast_comps(comps, datafl, null=None):
    '''Returns a pandas DataFrame with upscaled composites
    and ASSUMES NO MISSING VALUES

    Parameters:
        datafl (DataFile): pygeostat DataFile with dh, ifrom, ito and at least 1 variable to upscale
        comps (DataFrame): pandas DataFrame with datafl.dh, datafl.ifrom, datafl.ito
        corresponding to the composite locations

    Keyword Args:
        null: the null value (values less than or equal to this will not be used)

    Returns:
        upscaled (DataFrame): Pandas DataFrame with value of upscaled variable

    .. codeauthor: pygeostat development team 2014-04-03'''

    if null is None:
        null = Parameters['data.null']
    # Get a list of drill holes for composites
    dhids = sorted(list(set(comps[datafl.dh])))
    # Determine the number of composites
    ncomps = len(comps[datafl.dh])
    # Create a dictionary of variable upscaling methods if not supplied
    vartypedict = {}
    for var in datafl.data.columns:
        if var not in [datafl.dh, datafl.ifrom, datafl.ito]:
            vartypedict[var] = 'continuous'
    # Determine number of variables
    nvar = len(vartypedict)
    # Initialize a pandas DataFrame for upscaled values
    upscaled = pd.DataFrame(np.ones([ncomps, nvar]) * null, columns=list(vartypedict))
    upscaled[datafl.dh] = comps[datafl.dh]
    upscaled[datafl.ifrom] = comps[datafl.ifrom]
    upscaled[datafl.ito] = comps[datafl.ito]
    # Now upscale
    for rowidx, row in comps.iterrows():
        compfrom = row[datafl.ifrom]
        compto = row[datafl.ito]
        # Get the variable, froms and tos which are in that drill hole
        dhdata = datafl.data[datafl.data[datafl.dh] == row[datafl.dh]]
        # Define local function to get length in interval

        def length_in_interval(datarow):
            'Define local function to get length in interval'
            # Is the variable value valid?
            if datarow[var] <= null:
                return 0.0

            # Case 1 - entire sample interval in composite
            if (datarow[datafl.ifrom] >= compfrom) and (datarow[datafl.ito] <= compto):
                return datarow[datafl.ito] - datarow[datafl.ifrom]

            # Case 2 - partially bottom of interval
            elif ((datarow[datafl.ifrom] >= compfrom) and (datarow[datafl.ifrom] < compto) and
                  (datarow[datafl.ito] > compto)):

                return compto - datarow[datafl.ifrom]

            # Case 3 - partially in top of interval
            elif ((datarow[datafl.ifrom] < compfrom) and (datarow[datafl.ito] > compfrom) and
                  (datarow[datafl.ito] <= compto)):

                return datarow[datafl.ito] - compfrom

            # Case 4 - composite bracketed by sample interval
            elif (datarow[datafl.ifrom] <= compfrom) and (datarow[datafl.ito] >= compto):
                return compto - compfrom

            # Case 5 - no length in interval
            else:
                return 0.0
        # Get length of each sample in composite
        dhdata['Length'] = dhdata.apply(length_in_interval, axis=1)
        # If no data was found, return null value
        if sum(dhdata['Length']) <= 0.0:
            for var, vartype in vartypedict.items():
                upscaled.loc[rowidx, var] = null
        # Upscale the data
        for var, vartype in vartypedict.items():
            upscaled.loc[rowidx, var] = np.dot(dhdata[var],
                                               dhdata['Length']) / sum(dhdata['Length'])
    return upscaled


def write_vtp(datafl, vartypes, drillholes, outflname, complength=5.0, null=None):
    '''Generates a VTP file compatible with ParaView.

    Parameters:
        datafl (DataFile): pygeostat DataFile with dh, ifrom, ito and at least 1 variable to upscale
        vartypes (dict): dictionary of variable types like {'Grade':'continuous', 'Category':'categorical'}
        drillholes (dict): dictionary of drillhole IDs to pygeostat DrillHole objects
                           likely obtained by running:
                           drillholes = gs.set_desurvey(collarfl, surveyfl, 'Depth', 'Azimuth', 'Inclination')
        outflname (str): output VTP file to generate

    Keyword Args:
        null: the null value (values equal to this will not be used)
        complength: this is the composite length which everything is regularized to which helps
                    prevent visual artefacts and long triangles with a tube filter

    .. codeauthor: pygeostat development team 2017'''
    from ..datautils import fastcomps as gsfastcomps
    if null is None:
        null = Parameters['data.null']

    # "Composite" it down to reduce visual artefacts
    tmpfl = copy.deepcopy(datafl)
    comps = set_comps(datafl, complength)
    tmpfl.data = gsfastcomps.get_comps(comps, datafl, vartypes=vartypes, null=null)

    # Slice out any assays exceeding maximum hole depth
    tmpdfs = []
    for dhid, drillhole in drillholes.iteritems():
        tmpdfs.append(tmpfl.data[(tmpfl.data[tmpfl.dh] == dhid) &
                                 (tmpfl.data[tmpfl.ifrom] <= max(drillhole.data['Along'])) &
                                 (tmpfl.data[tmpfl.ito] <= max(drillhole.data['Along']))])
    tmpfl.data = pd.concat(tmpdfs, ignore_index=False)

    # Desurvey in all permutations...
    # Mid desurveying
    get_desurvey(tmpfl, drillholes, inplace=True, x='MIDX', y='MIDY', z='MIDZ')
    # Top desurveying
    tmp_name = tmpfl.ito
    tmpfl.ito = None
    get_desurvey(tmpfl, drillholes, inplace=True, x='TOPX', y='TOPY', z='TOPZ')
    # Bottom desurveying
    tmpfl.ito = tmp_name
    tmp_name = tmpfl.ifrom
    tmpfl.ifrom = None
    get_desurvey(tmpfl, drillholes, inplace=True, x='BOTTOMX', y='BOTTOMY', z='BOTTOMZ')
    tmpfl.ifrom = tmp_name

    # Now generate the VTK file
    root = ET.Element('VTKFile', type="PolyData", version="1.0", byte_order="LittleEndian", header_type="UInt64")
    pdata = ET.SubElement(root, 'PolyData')

    # Each drillhole is a separate "Piece" of "PolyData" in the VTK File
    for dhid, drillhole in drillholes.iteritems():
        dhdf = tmpfl.data[tmpfl.data[tmpfl.dh] == dhid]
        nverts = len(dhdf) + 2 # Add 2 since we have the top and bottom extra vertices

        # Header
        piece = ET.SubElement(pdata, 'Piece', NumberOfPoints="{}".format(nverts), NumberOfVerts="0",
                              NumberOfLines="1", NumberOfStrips="0", NumberOfPolys="0")

        # Vertex coordinates
        points = ET.SubElement(piece, 'Points')
        darray = ET.SubElement(points, 'DataArray', type="Float32", Name="Points",
                               NumberOfComponents="3", format="ascii")

        # Top and bottom vertices are added to make the drill hole match the true length
        ptarray = ' '.join(map(str, dhdf[['TOPX', 'TOPY', 'TOPZ']].values[0]))
        ptarray = ptarray + ' ' + ' '.join([' '.join(map(str, row[['MIDX', 'MIDY', 'MIDZ']].values)) for \
                                      rowidx, row in dhdf.iterrows()])
        ptarray = ptarray + ' ' + ' '.join(map(str, dhdf[['BOTTOMX', 'BOTTOMY', 'BOTTOMZ']].values[-1]))
        darray.text = ptarray

        # Variable values
        pointdata = ET.SubElement(piece, 'PointData', Scalars="DATA")
        for varname, vartype in vartypes.iteritems():
            if vartype.lower().startswith('cat'):
                vtktype = 'Int32'
                ptarray = str(int(dhdf[varname].values[0])) + ' ' + \
                          ' ' .join(map(str, map(int, dhdf[varname].values))) + ' ' + \
                          str(int(dhdf[varname].values[-1]))
            else:
                vtktype = 'Float32'
                ptarray = str(dhdf[varname].values[0]) + ' ' + \
                          ' ' .join(map(str, dhdf[varname].values)) + ' ' + \
                          str(dhdf[varname].values[-1])

            darray = ET.SubElement(pointdata, 'DataArray', type=vtktype, Name=varname, format="ascii")
            darray.text = ptarray

        # Line values
        lines = ET.SubElement(piece, 'Lines')
        darray = ET.SubElement(lines, 'DataArray', type="Int64", Name="connectivity", format="ascii")
        darray.text = ' '.join(map(str, range(0,nverts)))
        darray = ET.SubElement(lines, 'DataArray', type="Int64", Name="offsets", format="ascii")
        darray.text = '{}'.format(nverts)

    tree = ET.ElementTree(root)
    tree.write(outflname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''utils.py: Contains utility functions for providing status alerts in a script'''
from __future__ import absolute_import, division, print_function
__author__ = 'pygeostat development team'
__date__ = '2015'
__version__ = '1.000'

import textwrap


def printerr(text, width=80, errtype=None):
    """
    Small utility to print custom errors with proper indentation and text wrapping. The only error
    types coded are `error` and `warning`.

    Parameters:
        text (str): String of text without the preceding error flag that will be formated
        width (int): The maximum length of wrapped lines
        errtype (str): Indicate which type of error to format the string as. The default value of
            `None` will only text wrap the string to the specified `width`.

    .. codeauthor:: pygeostat development team 2015-11-01
    """
    import pygeostat as gs

    if isinstance(errtype, str):
        errtype.lower()
    if errtype is 'error':
        text = 'ERROR: ' + text
        subsequent_indent = "       "
    elif errtype is 'warning':
        text = 'WARNING: ' + text
        subsequent_indent = "         "
    else:
        subsequent_indent = ""
    print(textwrap.fill(text, width=width, subsequent_indent=subsequent_indent))


def log_progress(sequence, size=None, name='Items'):
    """
    Parameters:
        sequence: The thing that is being iterated over

    Example:
        >>> from pygeostat import log_progress
        >>> from time import sleep
        >>> for i in log_progress(range(200)):
        ...     sleep(0.1)

    .. image:: ./figures/log_progress.gif

    """
    try:
        from tqdm import tqdm, tqdm_notebook
    except:
        print("pip install tqdm to install the progressbar functionality !")
    try:
        get_ipython()
        return tqdm_notebook(sequence, desc=name, total=size)
    except:
        return tqdm(sequence, name, total=size, ascii=True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Generate a loadings plot depicting the loadings or correlation between the original variables and
their transformed counterparts."""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt

from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters


@set_plot_style
def loadings_plot(loadmat, figsize=None, ax=None, title=None, xticklabels=None, yticklabels=None,
                rotateticks=None, plot_style=None, custom_style=None, output_file=None, **kwargs):
    """
    This function uses matplotlib to create a loadings plot with variably sized colour mapped
    boxes illustrating the contribution of each of the input variables to the transformed variables.

    The only parameter needed ``loadmat`` containing the loadings or correlation matrix. All of the
    other arguments are optional. Figure size will likely have to be manually adjusted. If
    ``xticklabels`` and/or ``yticklabels`` are left to their default value of ``None`` and the input
    matrix is contained in a pandas dataframe, the index/column information will be used to label
    the columns and rows. If a numpy array is passed, axis tick labels will need to be provided.
    Axis tick labels are automatically checked for overlap and if needed, are rotated. If rotation
    is necessary, consider condensing the variable names or plotting a larger figure if the result
    appears odd.

    Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    Parameters:
        loadmat: Pandas dataframe or numpy matrix containing the required loadings or correlation
            matrix
        figsize (tuple): Figure size (width, height)
        ax (mpl.axis): Matplotlib axis to plot the figure
        title (str): Title for the plot.
        xticklabels (list): Tick labels along the x-axis
        yticklabels (list): Tick labels along the y-axis
        rotateticks (bool tuple): Indicate if the axis tick labels should be rotated (x, y)
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        **kwargs: Optional permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`

    Returns:
        ax (ax): matplotlib Axes object with the loadings plot


    Examples:

	Grab the correlation between the PCA variables and their corresponding input variables as
	a pandas dataframe:

	.. plot::

		import pygeostat as gs
		data_file = gs.ExampleData('3d_correlation')
		loadmat = data_file.data.corr().iloc[3:6,6:9] 
		gs.loadings_plot(loadmat.values, figsize=(5,5), xticklabels=['PC1', 'PC2', 'PC3'], yticklabels=['InputVariable1', 'InputVariable2', 'InputVariable3']) 

    """
    from .utils import _tickoverlap
    from .export_image import export_image

    # Set-up plot if no axis is supplied
    if ax is None:
        _, ax = plt.subplots(1, figsize=figsize)
    ax.set_aspect('equal')
    # Set title if required
    if title:
        ax.set_title(title)
    # Set the labels if possible
    if xticklabels and not isinstance(xticklabels, bool):
        xlabels = xticklabels
    elif hasattr(loadmat, 'columns'):
        xlabels = list(loadmat.columns)
    else:
        xlabels = None
    if yticklabels and not isinstance(yticklabels, bool):
        ylabels = yticklabels
    elif hasattr(loadmat, 'index'):
        ylabels = list(loadmat.index)
    else:
        ylabels = None
    # Set-up some parameters
    nx = loadmat.shape[1]
    ny = loadmat.shape[0]
    # Set-up colour mapping
    clrnorm = mpl.colors.Normalize(vmin=-1, vmax=1)
    clrmap = mpl.cm.ScalarMappable(norm=clrnorm, cmap='bwr')
    # Set-up axis parameters
    ax.set(xlim=(0, nx), ylim=(0, ny))
    xticklocs = np.arange(nx)
    yticklocs = np.arange(ny)
    # Set-up x axis labels and grid
    ax.xaxis.tick_top()
    ax.set_xticks(xticklocs + 0.5)
    ax.set_xticks(xticklocs, minor=True)
    if xlabels is not None:
        ax.set_xticklabels(xlabels, va='bottom', ha='center', rotation='horizontal')
        ax.tick_params(axis='x', pad=2)
    else:
        ax.get_xaxis().set_ticks([])
    # Set-up y axis labels and grid
    ax.invert_yaxis()
    ax.set_yticks(yticklocs + 0.5)
    ax.set_yticks(yticklocs, minor=True)
    if ylabels is not None:
        ax.set_yticklabels(ylabels, va='center', ha='right', rotation='vertical')
        ax.tick_params(axis='y', pad=2)
    else:
        ax.get_yaxis().set_ticks([])
    # The plots tick labels will not be properly accessible until the figure is "drawn", once the
    # command below is run, ax.get_ticklabel() will actually work properly.
    plt.draw()
    # Check if the axis tick labels overlap, if so rotate them
    if rotateticks is None:
        rotateticks = Parameters['plotting.rotateticks']
    if rotateticks is None:
        rotateticks = _tickoverlap(ax)
    # Rotate if required
    if rotateticks[0]:
        xlabels = ax.get_xticklabels()
        for xlabel in xlabels:
            xlabel.set_ha('center')
            xlabel.set_va('bottom')
            xlabel.set_rotation(45)
    if rotateticks[1]:
        ylabels = ax.get_yticklabels()
        for ylabel in ylabels:
            ylabel.set_ha('right')
            ylabel.set_va('center')
            ylabel.set_rotation(-45)
    # Plot grid
    plt.grid(False)
    plt.grid(True, which='minor')
    # Convert pandas dataframe to a numpy matrix
    if isinstance(loadmat, pd.DataFrame):
        loadmat = loadmat.values
    # Plot the loading blocks
    for y in range(ny):
        for x in range(nx):
            corr = loadmat[y, x]
            length = abs(corr) * 0.8
            boxstart = (0.5 - (length / 2))
            rectangle = mpl.patches.Rectangle((boxstart + x, (y + 0.15)), length, 0.35)
            rectangle.set(edgecolor='grey', facecolor=clrmap.to_rgba(corr))
            ax.annotate(round(corr, 2), xy=(x + .5, y + 0.8), ha='center')
            ax.add_patch(rectangle)
    # Export figure
    if output_file is not None:
        export_image(output_file, **kwargs)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''labels.py: Contains utilities for creating index columns and labels'''
from __future__ import absolute_import, division, print_function
__author__ = 'Tyler Acorn'
__date__ = '2015-10-08'
__version__ = '1.001'

import numpy as np


def insert_real_idx(data, num_real=0, bindex=True, real_column='Realization',
                    bi_column='BlockIndex'):
    '''
    This will insert realization index columns. By default it will use the griddef associated
    with the file.

    Parameters:
        num_real (int): If you do not have a griddef associated with the file you can tell it how
            many realizations there are
        bindex (bool): True or False for adding a block index
        real_column (str): Set the name of the column used for the Realizations Index
        bi_column (str): Set the name of the column used for the Block Index

    **Process**

        If there are already a "real_column" or "bi_column" columns it will overwrite the values in
        these columns If the "real_column" and "bi_column" columns aren't in the dataframe it will
        insert these columns at the front.

    .. codeauthor:: pygeostat development team 2015-09-30
    '''

    from ..data import DataFile
    # ---------------
    # Insert columns at front of the dataframe if they do not already exist
    # ---------------
    if real_column not in data.data.columns:
        data.data.insert(loc=0, column=real_column, value=0)
        if bindex is True:
            if bi_column not in data.columns and bindex is True:
                data.data.insert(loc=1, column=bi_column, value=0)
    # ---------------
    # Calculate Number of Realizations and the size of the realizations
    # ---------------
    if data.griddef is None:
        if num_real is not 0:
            real_size = len(data.data.index) / num_real
            # The next step I make sure we have enough rows in our dataframe. is this necessary?
    elif data.griddef is not None:
        num_real = len(data.data.index) / data.griddef.count()
        real_size = data.griddef.count()
    else:
        print('ERROR: No griddef assoicated with file and num_real is still 0')
    temp_labels = make_labels('R', (int(num_real)), padding=(len(str(int(num_real)))))
    ridx = np.repeat(temp_labels, real_size)
    data.data[real_column] = ridx
    if bindex is True:
        temp_labels = make_labels('B', (int(real_size)), padding=(len(str(int(real_size)))))
        bidx = np.tile(temp_labels, num_real)
        data.data[bi_column] = bidx


def make_labels(prefix, num, padding=0):
    '''
    Returns a series of lables combining a prefix and a number with leading zeros

    Parameters:
        prefix (str): any letter(s) that you want as the prefix (for example B for blockindex)
        num (int): The number of labels you want.
        padding (int): if given an integer value will pad the numbers with zeros until the prefix +
            the numbers equal the length of the padding value

    Return:
        Series: This will return a series with "n" number of labels starting from 1

    Note:
        Barrowed from website http://pandas.pydata.org/pandas-docs/stable/advanced.html#advanced

    Examples:
        Creating an array of labels

        >>> label = gs.datautils.make_labels('R', 3, padding=3)
        >>> label
        >>> [R001, R002, R003]

    .. codeauthor:: pygeostat development team 2015-09-21
    '''

    return ["%s%s" % (prefix, str(i).zfill(padding)) for i in range(1, (num + 1))]



'''
The :mod:`pygslib.programs` module is for integrating with gslib fortran programs.
'''
from __future__ import absolute_import, division, print_function

from .programs import *
from .program_utils import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A basic histogram plotting routine using matplotlib"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import scipy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters

@set_plot_style
def histogram_plot(data, var=None, weights=None, cat=None, catdict=None, bins=None, icdf=False, lower=None,
                    upper=None, ax=None, figsize=None, xlim=None, ylim=None, title=None, xlabel=None,
                    stat_blk=None, stat_xy=None, stat_ha=None, roundstats=None, sigfigs=None, color=None,
                    edgecolor=None, edgeweights=None, grid=None, axis_xy=None, label_count=False,
                    rotateticks=None, plot_style=None, custom_style=None, output_file=None, out_kws=None,
                    stat_fontsize=None, stat_linespacing=None, logx=False, **kwargs):
    """
    Generates a matplotlib style histogram with summary statistics. Trimming is now only applied
    to NaN values (Pygeostat null standard).

    The only required required parameter is ``data``. If ``xlabel`` is left to its default value of
    ``None`` and the input data is contained in a pandas dataframe or series, the column
    information will be used to label the x-axis.

    Two statistics block sets are available: ``'all'`` and the default ``'minimal'``. The
    statistics block can be customized to a user defined list and order. Available statistics are
    as follows:

    >>> ['count', 'mean', 'stdev', 'cvar', 'max', 'upquart', 'median', 'lowquart', 'min',
    ...  'p10', 'p90']

    The way in which the values within the statistics block are rounded and displayed can be
    controlled using the parameters ``roundstats`` and ``sigfigs``.

    Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    Parameters:
        data (np.ndarray, pd.DataFrame/Series, or gs.DataFile): data array, which must be 1D
            unless var is provided. The exception being a DataFile, if data.variables
            is a single name.
        var (str): name of the variable in data, which is required if data is not 1D.
        weights (np.ndarray, pd.DataFrame/Series, or gs.DataFile or str): 1D array of declustering
             weights for the data. Alternatively the declustering weights name in var. If data
             is a DataFile, it may be string in data.columns, or True to use data.weights
             (if data.weights is not None).
        cat (bool or str): either a cat column in data.data, or if True uses data.cat if data.cat
            is not None
        catdict (dict or bool): overrides bins. If a categorical variable is being plotted, provide
            a dictionary where keys are numeric (categorical codes) and values are their associated
            labels (categorical names). The bins will be set so that the left edge (and associated
            label) of each bar is inclusive to each category. May also be set to True, if data is
            a DataFile and data.catdict is initialized.
        bins (int or list): Number of bins to use, or a list of bins
        icdf (bool): Indicator to plot a CDF or not
        lower (float): Lower limit for histogram
        upper (float): Upper limit for histogram
        ax (mpl.axis): Matplotlib axis to plot the figure
        figsize (tuple): Figure size (width, height)
        xlim (float tuple): Minimum and maximum limits of data along the x axis
        ylim (float tuple): Minimum and maximum limits of data along the y axis
        title (str): Title for the plot
        xlabel (str): X-axis label
        stat_blk (bool): Indicate if statistics are plotted or not
        stat_xy (float tuple): X, Y coordinates of the annotated statistics in figure
            space. Based on Parameters['plotting.histogram_plot.stat_xy'] if a histogram and
            Parameters['plotting.histogram_plot.stat_xy'] if a CDF, which defaults to the top right when
            a PDF is plotted and the bottom right if a CDF is plotted.
        stat_ha (str): Horizontal alignment parameter for the annotated statistics. Can be
            ``'right'``, ``'left'``, or ``'center'``. If None, based on
            Parameters['plotting.stat_ha']
        stat_fontsize (float): the fontsize for the statistics block. If None, based on
            Parameters['plotting.stat_fontsize']. If less than 1, it is the fraction of the
            matplotlib.rcParams['font.size']. If greater than 1, it the absolute font size.
        roundstats (bool): Indicate if the statistics should be rounded to the number of digits or
            to a number of significant figures (e.g., 0.000 vs. 1.14e-5). The number of digits or
            figures used is set by the parameter ``sigfigs``. sigfigs (int): Number of significant
            figures or number of digits (depending on ``roundstats``) to display for the float
            statistics
        color (str or int or dict): Any permissible matplotlib color or a integer which is used to draw
            a color from the pygeostat color pallet ``pallet_pastel``> May also be a dictionary of colors,
            which are used for each bar (useful for categories). colors.keys() must align with bins[:-1]
            if a dictionary is passed. Drawn from Parameters['plotting.cmap_cat'] if catdict is used
            and their keys align.
        edgecolor (str): Any permissible matplotlib color for the edge of a histogram bar
        grid(bool): plots the major grid lines if True. Based on Parameters['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
        label_count (bool): label the number of samples found for each category in catdict. Does
            nothing if no catdict is found
        rotateticks (bool tuple): Indicate if the axis tick labels should be rotated (x, y)
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        **kwargs: Optional permissible keyword arguments to pass to either: (1) matplotlib's hist
            function if a PDF is plotted or (2) matplotlib's plot function if a CDF is plotted.

    Returns:
        ax (ax): matplotlib Axes object with the histogram

    **Examples:**

    A simple call:

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        # plot the histogram_plot
        gs.histogram_plot(dfl, var="Phi", bins=30)

    |

    Change the colour, number of significant figures displayed in the statistics, and pass some
    keyword arguments to matplotlibs hist function:

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        # plot the histogram_plot
        gs.histogram_plot(dfl, var="Phi", color='#c2e1e5', sigfigs=5, log=True, density=True)

    |

    Plot a CDF while also displaying all available statistics, which have been shifted up:

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        # plot the histogram_plot
        gs.histogram_plot(dfl, var="Phi", icdf=True, stat_blk='all', stat_xy=(1, 0.75))
        # Change the CDF line colour by grabbing the 3rd colour from the color pallet
        # ``cat_vibrant`` and increase its width by passing a keyword argument to matplotlib's
        # plot function. Also define a custom statistics block:
        gs.histogram_plot(dfl, var="Phi", icdf=True, color=3, lw=3.5, stat_blk=['count','upquart'])

    |

    Generate histograms of Phi considering the categories:

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        cats = [1, 2, 3, 4, 5]
        colors = gs.catcmapfromcontinuous("Spectral", 5).colors
        # build the required cat dictionaries
        dfl.catdict = {c: "RT {:02d}".format(c) for c in cats}
        colordict =  {c: colors[i] for i, c in enumerate(cats)}
        # plot the histogram_plot
        f, axs = plt.subplots(2, 1, figsize=(8, 6))
        for var, ax in zip(["Phi", "Sw"], axs):
            gs.histogram_plot(dfl, var=var, cat=True, color=colordict, bins=40, figsize=(8, 4), ax=ax,
                       xlabel=False, title=var)

    |

    Generate cdf subplots considering the categories:

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        cats = [1, 2, 3, 4, 5]
        colors = gs.catcmapfromcontinuous("Spectral", 5).colors
        # build the required cat dictionaries
        dfl.catdict = {c: "RT {:02d}".format(c) for c in cats}
        colordict =  {c: colors[i] for i, c in enumerate(cats)}
        # plot the histogram_plot
        f, axs = plt.subplots(2, 2, figsize=(12, 9))
        axs=axs.flatten()
        for var, ax in zip(dfl.variables, axs):
            gs.histogram_plot(dfl, var=var, icdf=True, cat=True, color=colordict, ax=ax)

    Recreate the `Proportion` class plot

    .. plot::

        import pygeostat as gs
        # load some data
        dfl = gs.ExampleData("point3d_ind_mv")
        cats = [1, 2, 3, 4, 5]
        colors = gs.catcmapfromcontinuous("Spectral", 5).colors
        # build the required cat dictionaries
        dfl.catdict = {c: "RT {:02d}".format(c) for c in cats}
        colordict =  {c: colors[i] for i, c in enumerate(cats)}
        # plot the histogram_plot
        ax = gs.histogram_plot(dfl, cat=True, color=colordict, figsize=(7, 4), rotateticks=(45, 0),
                        label_count=True)

    """
    import pygeostat as gs
    from . utils import format_plot, _set_stat_fontsize, _format_grid, _format_tick_labels, setup_plot, catcmapfromcontinuous
    from . cmaps import _cat_pastel_data, _cat_vibrant_data
    import copy
    # Now converting to a numpy array, as encountering some odd pandas performance, and there's
    # no major disadvantagve to application of a numpy in this context to my knowledge - RMB
    # If a list is passed convert it to a series so that trimming can take place
    # weights
    if isinstance(weights, str):
        if isinstance(data, pd.DataFrame) or isinstance(data, gs.DataFile):
            weights = data[weights]
    elif isinstance(weights, bool):
        if weights:
            if isinstance(data, gs.DataFile):
                if data.weights is None:
                    raise ValueError('weights=True but data.weights is None!')
                elif isinstance(data.weights, list):
                    raise ValueError('weights=True but data.weights is a list!')
                weights = data[data.weights].values
            else:
                raise ValueError('weights=True is only valid if data is a DataFile!')
        else:
            weights = None
    if isinstance(weights, pd.Series) or isinstance(weights, pd.DataFrame):
        weights = weights.values
    # cats for continuous histogram_plots
    if isinstance(cat, str):
        if isinstance(data, pd.DataFrame) or isinstance(data, gs.DataFile):
            cat = data[cat]
    elif isinstance(cat, bool):
        if cat:
            if isinstance(data, gs.DataFile):
                if data.cat is None:
                    raise ValueError('cat=True but data.cat is None!')
                cat = data[data.cat].values
                if catdict is None and data.catdict is None:
                    raise ValueError("pass a `catdict` when setting `cat`")
                else:
                    catdict = data.catdict
            else:
                raise ValueError('cat=True is only valid if data is a DataFile!')
        else:
            cat = None
    if isinstance(cat, pd.Series) or isinstance(cat, pd.DataFrame):
        cat = cat.values
    # Handle categorical dictionary
    if isinstance(catdict, bool):
        if catdict:
            if not isinstance(data, gs.DataFile):
                raise ValueError('catdict as a bool is only valid if data is a DataFile!')
            if data.catdict is None:
                raise ValueError('catdict as a bool is only valid if data is not None!')
            catdict = data.catdict
    # Variable
    # Handle data that is 2-D and/or a DataFile
    if isinstance(var, str):
        if isinstance(data, pd.DataFrame) or isinstance(data, gs.DataFile):
            if isinstance(cat, str):
                cat = data[cat]
            data = data[var]
        else:
            raise ValueError('var as a string is only valid if data is a DataFile or DataFrame!')
    elif isinstance(data, gs.DataFile):
        if isinstance(data.variables, str):
            data = data[data.variables]
        elif cat is not None:
            if isinstance(cat, str):
                data = data[cat]
            elif var is None and isinstance(cat, (np.ndarray, list)):
                data = cat
        elif len(data.columns) == 1:
            data = data.data
        else:
            raise ValueError('Could not coerce data (DataFile) into a 1D dataset!')
    # Get the xlabel if possible before converting to a numpy array
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        if xlabel is None:
            xlabel = gs.get_label(data)
        data = data.values
    elif isinstance(data, list):
        data = np.array(data)
    if isinstance(cat, (pd.Series, pd.DataFrame)):
        cat = cat.values
    # Should be numpy by now...
    if data.ndim > 1:
        if data.shape[1] > 1:
            raise ValueError('Could not coerce data into a 1D dataset!')
        else:
            data = data.flatten()
    # Handle Null values if needed
    idx = np.isnan(data)
    nullcnt = np.sum(idx)
    if nullcnt > 0:
        data = data[~idx]
        if weights is not None:
            weights = weights[~idx]
        if cat is not None:
            cat = cat[~idx]
    # Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()
    # Set-up plot if no axis is supplied
    _, ax, _ = setup_plot(ax, figsize=figsize, aspect=False)
    # Infer some default parameters
    if weights is None:
        weights = np.ones(len(data)) / len(data)
    else:
        weights = weights / np.sum(weights)
    # Some quick error checks
    assert(np.all(weights) >= 0.0), 'weights less than 0 not valid'
    # Categories
    if isinstance(catdict, dict) and var is None:
        if not all([isinstance(float(i), float) for i in catdict.keys()]):
            raise ValueError('if catdict is dict., all keys should be an int/float!')
        # The bins are set to begin at the start of each category
        # bins go from 0.5 to (icat + 1) + 0.5
        # label is centered at (icat + 1)
        bins = np.arange(len(catdict) + 1) + 0.5
    if color is None and isinstance(catdict, dict):
        # Color each bin by the category color?
        if isinstance(Parameters['plotting.cmap_cat'], dict):
            temp = Parameters['plotting.cmap_cat']
            if list(sorted(temp.keys())) == list(sorted(catdict.keys())):
                color = temp
        else:
            color = catcmapfromcontinuous(Parameters["plotting.cmap"], len(catdict)).colors
    if isinstance(color, dict):
        if list(sorted(color.keys())) != list(sorted(catdict.keys())):
            raise ValueError(('if color is a dictionary, keys must align with '
                              'bins[:-1]! Consider using a single color.'))
        temp = color
        color = []
        for _, v in sorted(temp.items()):
            color.append(v)
    # Color setup
    if isinstance(color, int):
        # Grab a color from ``cat_vibrant`` if an integer is passed
        color = _cat_pastel_data[color % len(_cat_vibrant_data)]
    if not icdf:
        if color is None:
            color = Parameters['plotting.histogram_plot.facecolor']
        if edgecolor is None:
            edgecolor = Parameters['plotting.histogram_plot.edgecolor']
        if edgeweights is None:
            if "lw" in kwargs:
                edgeweights = kwargs.pop("lw")
            else:
                edgeweights = Parameters["plotting.histogram_plot.edgeweight"]
    else:
        if color is None and icdf:
            color = Parameters['plotting.histogram_plot.cdfcolor']
    plotdata = copy.deepcopy(data)
    plotweights = copy.deepcopy(weights)
    if xlim is not None:
        plotdata[data < xlim[0]] = xlim[0]
        plotdata[data > xlim[1]] = xlim[1]
    # Main plotting
    if icdf:

        def singlecdf(ax, data, weights, lower, upper, bins, color, label=None, **kwargs):
            """ local function to plot a single cdf """
            cdf_x, cdfvals = gs.cdf(data, weights=weights, lower=lower, upper=upper, bins=bins)
            # Matplotlib is a memory hog if to many points are used. Limit the number of points the CDF
            # is build with to 1000. The tails are given extra attention to make sure they are defined
            # nicely.
            if len(cdf_x) > 1000:
                cdfinterp = scipy.interpolate.interp1d(x=cdfvals, y=cdf_x)
                cdfvals = np.concatenate([np.arange(cdfvals.min(), 0.1, 0.001),
                                          np.arange(0.1, 0.9, 0.01),
                                          np.arange(0.9, cdfvals.max(), 0.001)])
                cdf_x = []
                for val in cdfvals:
                    cdf_x.append(cdfinterp(val))
                cdf_x = np.array(cdf_x)
            fig = ax.plot(cdf_x, cdfvals, color=color, label=label, **kwargs)
            return fig

        if catdict is not None:
            if var is not None:
                stat_blk = False
                for icat, c in enumerate(catdict):
                    clr = color[icat]
                    catidx = cat == c
                    fig = singlecdf(ax, plotdata[catidx], plotweights[catidx], lower, upper, bins, clr,
                                    label=catdict[c], **kwargs)
            else:
                raise ValueError("`icdf=True` and `catdict` only makes sense with a `var` defined")
        else:
            fig = singlecdf(ax, plotdata, plotweights, lower, upper, bins, color, **kwargs)
        if ylim is None:
            ylim = (0, 1.0)
    else:
        if bins is None:
            bins = Parameters['plotting.histogram_plot.histbins']
        label = kwargs.pop("label", None)
        if bins is None:
            if len(plotdata) < 200:
                bins = 20
            elif len(plotdata) < 500:
                bins = 25
            else:
                bins = 30
        if logx:
            if catdict is not None:
                raise ValueError('Cannot have logx with catdict!')
            if xlim is None:
                minv = np.log10(max(plotdata.min(), 1e-10))
                maxv = np.log10(plotdata.max())
            else:
                minv = np.log10(max(xlim[0], 1e-10))
                maxv = np.log10(xlim[1])
            if np.isnan([minv, maxv]).any():
                raise ValueError('ERROR converting your data to log base! are there negatives?')
            bins = np.logspace(minv, maxv, bins)
        if catdict is not None:
            if var is None:
                for icat, cat in enumerate(catdict):
                    plotdata[data == cat] = icat + 1
                histclr = None
            else:
                # generate lists of data per cat
                plotdata = [plotdata[cat == c] for c in catdict]
                plotweights = [weights[cat == c] for c in catdict]
                label = list(catdict.values())
                histtype = kwargs.pop("histtype", "stepfilled")
                stat_blk = False
                if "stacked" not in kwargs:
                    kwargs["stacked"] = True
                histclr = color
        histtype = kwargs.pop("histtype", "bar")
        if not isinstance(color, list):
            ax.hist(plotdata, bins, weights=plotweights, color=color, edgecolor=edgecolor,
                    histtype=histtype, label=label, lw=edgeweights, **kwargs)
        else:
            _, _, patches = ax.hist(plotdata, bins, weights=plotweights, histtype=histtype,
                                    color=histclr, edgecolor=edgecolor,
                                    label=label, lw=edgeweights, **kwargs)
            try:
                for patch, clr in zip(patches, color):
                    patch.set_facecolor(clr)
            except (AttributeError, ValueError):
                pass
        if catdict is not None and label_count:
            nd = len(data)
            for icat, cat in enumerate(catdict):
                count = np.count_nonzero(data == cat)
                pcat = (weights * (data == cat).astype(float)).sum()
                ax.text(icat + 1, pcat, count, ha="center", va="bottom")
    # Summary stats
    if stat_blk is None:
        stat_blk = Parameters['plotting.histogram_plot.stat_blk']
    if stat_xy is None:
        if icdf:
            stat_xy = Parameters['plotting.histogram_plot.stat_xy_cdf']
        else:
            stat_xy = Parameters['plotting.histogram_plot.stat_xy']
    if stat_blk:
        if sigfigs is None:
            sigfigs = Parameters['plotting.sigfigs']
        if roundstats is None:
            roundstats = Parameters['plotting.roundstats']
        if stat_ha is None:
            stat_ha = Parameters['plotting.stat_ha']
        if stat_linespacing is None:
            stat_linespacing = Parameters['plotting.stat_linespacing']
        if stat_linespacing is None:
            stat_linespacing = 1.0
        # Force no bins and upper/lower for median
        cdf_x, cdfvals = gs.cdf(data, weights=weights)
        # Currently defined statistics, possible to add more quite simply
        if np.mean(data) == 0:
            cdata = float("nan")
        elif roundstats:
            cdata = round((np.std(data) / np.mean(data)), sigfigs)
        else:
            cdata = gs.round_sigfig((np.std(data) / np.mean(data)), sigfigs)
        if roundstats:
            mean = round(gs.weighted_mean(data, weights), sigfigs)
            median = round(gs.percentile_from_cdf(cdf_x, cdfvals, 50.0), sigfigs)
            stdev = round(np.sqrt(gs.weighted_variance(data, weights)), sigfigs)
            minval = round(np.min(data), sigfigs)
            maxval = round(np.max(data), sigfigs)
            upquart = round(np.percentile(data, 75), sigfigs)
            lowquart = round(np.percentile(data, 25), sigfigs)
            p10 = round(np.percentile(data, 10), sigfigs)
            p90 = round(np.percentile(data, 90), sigfigs)
        else:
            mean = gs.round_sigfig(gs.weighted_mean(data, weights), sigfigs)
            median = gs.round_sigfig(gs.percentile_from_cdf(cdf_x, cdfvals, 50.0), sigfigs)
            stdev = gs.round_sigfig(np.sqrt(gs.weighted_variance(data, weights)), sigfigs)
            minval = gs.round_sigfig(np.min(data), sigfigs)
            maxval = gs.round_sigfig(np.max(data), sigfigs)
            upquart = gs.round_sigfig(np.percentile(data, 75), sigfigs)
            lowquart = gs.round_sigfig(np.percentile(data, 25), sigfigs)
            p10 = gs.round_sigfig(np.percentile(data, 10), sigfigs)
            p90 = gs.round_sigfig(np.percentile(data, 90), sigfigs)
        statistics = {'mean': (r'$m = %g$' % mean),
                      'median': (r'$x_{{50}} = %g$' % median),
                      'count': ('$n = %i$' % len(data)),
                      'count_trimmed': ('$n_{trim} = %i$' % nullcnt),
                      'stdev': (r'$\sigma = %g$' % stdev),
                      'cvar': ('$CV = %g$' % cdata),
                      'min': ('$x_{{min}} = %g$' % minval),
                      'max': ('$x_{{max}} = %g$' % maxval),
                      'upquart': ('$x_{{75}} = %g$' % upquart),
                      'lowquart': ('$x_{{25}} = %g$' % lowquart),
                      'p10': ('$x_{{10}} = %g$' % p10),
                      'p90': ('$x_{{90}} = %g$' % p90)}
        # Default statistic sets
        if stat_blk == 'varlabel' and 'label' in kwargs:
            statistics['varlabel'] = kwargs['label']
        statsets = {'minimal': ['count', 'mean', 'median', 'stdev'],
                    'all': ['count', 'mean', 'stdev', 'cvar', 'max', 'upquart', 'median',
                            'lowquart', 'min'],
                    'varlabel': ['varlabel', 'count', 'mean', 'stdev', 'cvar', 'max', 'upquart',
                                 'median', 'lowquart', 'min'],
                    'none': None}
        # Use a default statistic set
        if isinstance(stat_blk, bool) and stat_blk:
            stat_blk = 'all'
        if isinstance(stat_blk, str):
            if stat_blk in statsets:
                stat_blk = statsets[stat_blk]
            else:
                print('WARNING: stats value of: "' + stat_blk + '" does not exist - '
                      'default to no stats')
                stat_blk = None
        # Use a supplied statistic set, but check for bad ones
        else:
            badstats = [s for s in stat_blk if s not in statistics]
            stat_blk = [s for s in stat_blk if s in statistics]
            for badstat in badstats:
                print('WARNING: stats value of: "' + badstat + '" does not exist - '
                      'It was removed from summary statistics list')
        # Form the stats string
        if stat_blk:
            if nullcnt != 0:
                stat_blk.insert(stat_blk.index('count') + 1, 'count_trimmed')
            stat_blk = [statistics[s] for s in stat_blk]
            txtstats = '\n'.join(stat_blk)
            if len(np.unique(weights)) > 1:
                txtstats = txtstats + '\n\nweights used'
            if stat_xy[1] > 0.5:
                va = 'top'
            else:
                va = 'bottom'
            # Set the stat_fontsize
            stat_fontsize = _set_stat_fontsize(stat_fontsize)
            ax.text(stat_xy[0], stat_xy[1], txtstats, va=va, ha=stat_ha, transform=ax.transAxes,
                    fontsize=stat_fontsize, linespacing=stat_linespacing)
    # Label as required
    if icdf:
        ylabel = 'Cumulative Distribution Function'
    elif 'density' in kwargs:
        ylabel = 'Probability Density Function (PDF)'
    else:
        ylabel = 'Frequency'
    ax = format_plot(ax, xlabel, ylabel, title, axis_xy=axis_xy, xlim=xlim, ylim=ylim, logx=logx)
    if catdict is not None and var is None:
        ticlocs = [i + 1 for i in range(len(catdict.keys()))]
        ax.set_xticks(ticlocs)
        ax.set_xticklabels(catdict.values())
        ax.set_xlim(0.25, len(catdict) + 0.75)
    elif catdict is not None and var is not None:
        ax.legend()
    _format_tick_labels(ax, rotateticks)
    # format_plot doesn't handle some specialized axis_xy and grid requirements
    # for histogram_plot...
    if icdf:
        # Ensure that we have top spline, in case it was removed above
        ax.spines['top'].set_visible(True)
        _format_grid(ax, grid, below=False)
    else:
        # The grid should be below for a histogram
        _format_grid(ax, grid, below=True)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        gs.export_image(output_file, **out_kws)
    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""drill_plot for a drill hole data set"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from . set_style import set_plot_style

@set_plot_style
def drill_plot(z, var, cat=None, categorical_dictionary=None, lw=2, line_color='k', barwidth=None, namelist=None,
			   legend_fontsize=10, title=None, ylabel=None, unit=None, grid=None, axis_xy=None,
			   reverse_y=False, xlim=None, ylim=None,  figsize=(2,10), ax=None,plot_style=None,
			   custom_style=None, output_file=None, out_kws=None, **kwargs):
	'''
	 A well log plot for both continuous and categorical variables. This plot handles one well log
	 plot at a time and the user can choose to generate subplots and pass the axes to this function
	 if multiple well log plots are required.

	Parameters:
		z (Elevation/Depth or distance along the well): Tidy (long-form) 1D data where a single
			column of the variable exists with each row is an observation. A pandas dataframe/series
			or numpy array can be passed.
		var (Variable being plotted): Tidy (long-form) 1D data where a single column of the variable
			exists with each row is an observation. A pandas dataframe/series or numpy array can be
			passed.
		lw (float): line width for log plot of a continuous variable
		line_color (string): line color for the continuous variable
		barwidth(float): width of categorical bars
		categorical_dictionary (dictionary): a dictionary of colors and names for categorical codes. E.g.
		{1: {'name': 'Sand', 'color': 'gold'},
		 2: {'name': 'SHIS','color': 'orange'},
		 3: {'name': 'MHIS','color': 'green'},
		 4: {'name': 'MUD','color': 'gray'}}
		legend_fontsize(float): fontsize for the legend plot rleated to the categorical codes. set
			this parameter to 0 if you do not want to have a legend
		title (str): title for the variable
		ylabel (str): Y-axis label, based on ``Parameters['plotting.zname']`` if None.
		unit (str): Unit to place inside the y-label parentheses, based on
			``Parameters['plotting.unit']`` if None.
		grid (bool): Plots the major grid lines if True. Based on ``Parameters['plotting.grid']`` if
			None.
		axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
			visible) if axis_xy is True. Based on ``Parameters['plotting.axis_xy']`` if None.
		reverse_y(bool): if true, the yaxis direction is set to reverse(applies to the cases that
			depth is plotted and not elevation)
		aspect (str): Set a permissible aspect ratio of the image to pass to matplotlib.
		xlim (float tuple): X-axis limits
		ylim (float tuple): Y-axis limits
		figsize (tuple): Figure size (width, height)
		ax (mpl.axis): Existing matplotlib axis to plot the figure onto
		out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`

	Returns:
		ax (ax): Matplotlib axis instance which contains the gridded figure

	Examples:

		A simple call to plot a continuous variable

	.. plot:: 

		import pygeostat as gs
		dat = gs.ExampleData('point3d_ind_mv')
		data = dat.data[dat.data['HoleID'] == 3] 
		gs.drill_plot(data['Elevation'], data['Sw'], grid = True)

	|

	Plot a categorical variable

	.. plot:: 

		import pygeostat as gs
		dat = gs.ExampleData('point3d_ind_mv')
		data = dat.data[dat.data['HoleID'] == 3] 
		gs.drill_plot(data['Elevation'], data['Lithofacies'])

	|

	Plot a categorical variable and provide a categorical dictionary
	
	.. plot:: 

		import pygeostat as gs
		dat = gs.ExampleData('point3d_ind_mv')
		data = dat.data[dat.data['HoleID'] == 3]
		cat_dict = {1: {'name': 'Sand', 'color': 'gold'},
			3: {'name': 'SHIS','color': 'orange'},
			4: {'name': 'MHIS','color': 'green'},
			5: {'name': 'MUD','color': 'gray'}}
		gs.drill_plot(data['Elevation'], data['Lithofacies'], categorical_dictionary=cat_dict)

	'''

	from matplotlib.ticker import FormatStrFormatter
	import matplotlib.patches as mpatches
	import matplotlib.pyplot as plt
	from .export_image import export_image
	import numpy as np
	import pandas as pd
	from .. pygeostat_parameters import Parameters
	from .utils import format_plot, get_label, setup_plot, catcmapfromcontinuous, _spatial_aspect

	# Sanity checks
	if not isinstance(z, (np.ndarray, np.generic, pd.Series)):
		raise ValueError("The parameter `z` must be a numpy array or a pandas series")
	if not isinstance(var, (np.ndarray, np.generic, pd.Series)):
		raise ValueError("The parameter `var` must be a numpy array or a pandas series")
	if title is None:
		title = get_label(var)
	if unit is None:
		unit = Parameters['plotting.unit']
	if unit is None or unit == '':
		unit = ''
	else:
		unit = ' ({})'.format(unit)
	if ylabel:
		ax.set_ylabel(ylabel)
	elif ylabel is None:
		ylabel = Parameters['plotting.zname'] + unit
	if not out_kws:
		out_kws = dict()

	fig, ax, cax = setup_plot(ax, figsize=figsize, aspect = False)

	# Create a new data frame
	df = pd.DataFrame()
	df['Z'] = np.array(z)
	df['Var'] = np.array(var)
	df.sort_values('Z', axis=0, ascending=True, inplace=True)
	df.reset_index(inplace=True, drop=True)


	if barwidth is None:
		barwidth = (np.max(df['Z'].values)  - np.min(df['Z'].values))/10
	
	if cat is None:
		if (len(set(df['Var'].values)) < (len(df) / 5)) :
			cat = True

	# Start of the plot
	if not cat:

		ax.plot(df['Var'], df['Z'], lw=lw, c=line_color, **kwargs)
		xlim = [np.nanmin(df['Var']), np.nanmax(df['Var'])]
		ylim = [np.nanmin(df['Z']), np.nanmax(df['Z'])]
		ax.set_xticks(np.linspace(np.nanmin(df['Var']),
								  np.nanmax(df['Var']), 2))
		ax.xaxis.set_major_formatter(FormatStrFormatter('%.1f'))
		ax.xaxis.set_label_position('top')
		ax.xaxis.set_ticks_position('top')
	else:
		categories = sorted(set(df['Var'].values))
		ncat = len(categories)
		
		# make sure grid option is not triggered
		grid = False

		if categorical_dictionary is None:
			colorlist = catcmapfromcontinuous("Spectral", ncat).colors
			namelist = []
			for i in range(ncat):
				string = 'Category Code %i' % (i + 1)
				namelist.append(string)
		
			categorical_dictionary = {}
			for i, item in enumerate(categories):
				categorical_dictionary.update({item: {'name': namelist[i], 'color': colorlist[i]} })
		else:
			if len(categorical_dictionary)<ncat:
				raise ValueError('The provided categorical_dictionary does not match the number of categorical codes')
			for item in categorical_dictionary.keys():
				if item not in categories:
					raise ValueError('The provided categorical code, {} does not exist in the provided data'.format(item))

		
		# First bar from bottom based on sorting the z column
		category = df['Var'][0]
		color = categorical_dictionary[category]['color']
		offset = (df['Z'][1] - df['Z'][0])/2
		z_b = df['Z'][0]
		z_t = z_b + offset
		ax.fill_between([0, barwidth], [z_b, z_b], [z_t,z_t], color = color)

		for i in range(1,len(df)-1):
			upper_offset = (df['Z'][i+1] - df['Z'][i])/2
			lower_offset = (df['Z'][i-1] - df['Z'][i])/2
			z_b = df['Z'][i] + lower_offset
			z_t = df['Z'][i] + upper_offset
			category = df['Var'][i]
			color = categorical_dictionary[category]['color']
			ax.fill_between([0, barwidth], [z_b, z_b], [z_t,z_t], color = color)

		# Last bar from bottom based on sorting the z column
		category = df['Var'][i+1]
		color = categorical_dictionary[category]['color']
		z_b = z_t
		z_t = df['Z'][i+1]
		ax.fill_between([0, barwidth], [z_b, z_b], [z_t,z_t], color = color)

		Patch_list = []
		for item in categorical_dictionary.values():
			Patch_list.append(mpatches.Patch(color=item['color'], label=item['name']))

		if legend_fontsize > 0:
			ax.legend(handles=Patch_list, loc='upper left',
					  bbox_to_anchor=(1, 1), fontsize=legend_fontsize)

		ylim = [np.nanmin(df['Z']), np.nanmax(df['Z'])]
		xlim = [0, barwidth]
		ax.xaxis.set_label_position('top')
		ax.get_xaxis().set_visible(False)

	format_plot(ax, xlabel='', ylabel = ylabel, title = title, grid = grid, axis_xy = axis_xy, xlim = xlim, ylim = ylim)
	
	if reverse_y:
		ax.invert_yaxis()

	if output_file or ('pdfpages' in out_kws):
		export_image(output_file, **out_kws)

	return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Basic variogram plotting function using matplotlib reminiscent of varplt from gslib"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import matplotlib as mpl
import matplotlib.pyplot as plt


import matplotlib as mpl
import matplotlib.pyplot as plt
from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters
import numpy as np



@set_plot_style
def variogram_plot(data, index=None, sill=1, experimental=True, label=None, ax=None, figsize=None,
		   xlim=None, ylim=None, title=None, xlabel=None, unit=None, ylabel=None, color=None,
		   marker=None, ms=None, ls=None, lw=None, minpairs=40, pairnumbers=False,
		   grid=None, axis_xy=None, plot_style=None, custom_style=None, output_file=None, out_kws=None, **kwargs):
	"""
	This function uses matplotlib to create a variogram plot. Input dataframe structure is
	important as the required data is found within columns that have recognizable headers.

	The only parameter needed is ``data`` and must be a pandas dataframe. All other arguments are
	optional or automatically determined. The settings for experimental and modeled variogram
	plotting is controlled by the ``experimental`` parameter.

	Please review the documentation of the :func:`gs.set_style()
	<pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
	<pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
	their use in this function can be understood.

	Parameters:
		data (pd.DataFrame/gs.DataFile): Dataframe/DataFile containing the variogram value, variogram distance, and
			variogram index (if required) data as columns. The dataframe must contain the correct
			column IDs. The column header containing the variogram distance can be: 'h', 'Lag
			Distance', or 'Distance.' The column header containing the variogram values can be:
			'vario', 'Variogram Value', or 'Variogram'
		index (int): Point to which variogram you would like to plot if there are multiple
			variogram within your dataframe. The dataframe must contain the correct column ID. The
			column header containing the variogram index values can be: 'Variogram Index' or
			'Index'
		sill (float): Value to plot a horizontal line representing the variograms sill
		experimental (bool): Indicator if the variogram is experimental ``True`` or modeled
			``False``
		label (str or bool): String to pass to Matplotlib's auto legend function. A default value
			will be generated; however, to prevent this, set label to ``False``
		ax (mpl.axis): Matplotlib axis to plot the figure
		figsize (tuple): Figure size (width, height)
		xlim (float tuple): Minimum and maximum limits of data along the x axis
		ylim (float tuple): Minimum and maximum limits of data along the y axis
		title (str): Title for the plot
		xlabel (str): X-axis label
		unit (str): Distance units used for lag distance. Only used if the keyword parameter
			``xlabel`` is left to its default value of ``None``.
		yalabl (str): Y-axis label
		color (str): Any Matplotlib color
		marker (str): A valid Matplotlib marker style
		ms (float): Marker size in points
		ls (float): A valid Matplotlib line style
		lw (float): Line width in points
		minpairs (int or bool): Any experimental variogram values that were calculated using fewer
			pairs then what is specified by the argument ``minpairs``, is highlighted red. To turn
			this functionality off, set ``minpairs`` to ``False``.
		grid (bool): Plots the major grid lines if True. Based on Parameters['plotting.grid']
			if None.
		axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
			visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
		plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
		output_file (str): Output figure file name and location
		out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
		**kwargs: Optional permissible keyword arguments to pass to matplotlib's plot function

	Returns:
		ax (ax): matplotlib Axes object with the variogram

	**Examples:**

		

		A simple call for experimental variograms, plotting only one direction:

		.. plot::
			
			import pygeostat as gs

			#Load the data from output from varcal/varmodel file
			varcalcdat = gs.ExampleData('experimental_variogram')
			gs.variogram_plot(varcalcdat, index=1)

		|

		A simple call for modeled variograms, plotting only one direction:

		.. plot::

			import pygeostat as gs
			
			varmodeldat = gs.ExampleData('variogram_model')
			gs.variogram_plot(varmodeldat, index=1, experimental=False)


		|

		Plot both experimental and modeled variograms for one direction:

		.. note::
			Some odd behavior may occur if the sill is repeatedly plotted. In the case when
			variograms are being plotted iteratively on the same figure, set the parameter ``sill``
			to ``False`` on all but the last figure.

		.. plot::

			import pygeostat as gs

			varcalcdat = gs.ExampleData('experimental_variogram')
			varmodeldat = gs.ExampleData('variogram_model')
			ax = gs.variogram_plot(varcalcdat.data, index=1, sill=False)
			gs.variogram_plot(varmodeldat.data, index=1, experimental=False, ax=ax)

		|

		Plot both directions experimental and modeled variograms with a legend, grab 2 colors
		from :func:`gs.get_palette() <pygeostat.plotting.utils.get_palette>` to use for the plots,
		and prevent points calculated using a low amount of pairs from being highlighted for one of
		the plots:

		.. plot::

			import pygeostat as gs

			varcalcdat = gs.ExampleData('experimental_variogram')
			varmodeldat = gs.ExampleData('variogram_model')

			colors = gs.get_palette('cat_dark', 2, cmap=False)
			ax = gs.variogram_plot(varcalcdat.data, index=1, color=colors[0], minpairs=False, label=False)
			gs.variogram_plot(varmodeldat.data, index=1, experimental=False, ax=ax, color=colors[0], label='Minor')
			gs.variogram_plot(varcalcdat.data, index=2, ax=ax, color=colors[1], label=False)
			gs.variogram_plot(varmodeldat.data, index=2, experimental=False, ax=ax, color=colors[1], label='Major')

			plt.legend(loc=4)

	"""
	from .export_image import export_image
	from ..data.data import DataFile
	from ..utility.logging import printerr
	from . utils import format_plot, setup_plot, _spatial_aspect

	
	if isinstance (data, DataFile):
		data = data.data

	# Handle dictionary defaults
	if out_kws is None:
		out_kws = dict()
	# Infer the data columns assuming pygeostat or gslib's varcalc function or pygeostat's model
	# function was used
	if 'h' in data.columns:
		h_name = 'h'
	elif 'Lag Distance' in data.columns:
		h_name = 'Lag Distance'
	elif 'Distance' in data.columns:
		h_name = 'Distance'
	else:
		print("Error: A distance column could not be found. Ensure its header is one of the",
			  "following: 'h',\n       'Lag Distance', or 'Distance.'")
		return
	if 'vario' in data.columns:
		vario_name = 'vario'
	elif 'Variogram Value' in data.columns:
		vario_name = 'Variogram Value'
	elif 'Variogram' in data.columns:
		vario_name = 'Variogram'
	else:
		print("Error: A variogram column could not be found. Ensure its header is one of the",
			  "\n       following: 'vario', 'Variogram Value', or 'Variogram.'")
		return
	if index:
		if 'Variogram Index' in data.columns:
			index_name = 'Variogram Index'
		elif 'Index' in data.columns:
			index_name = 'Index'
		else:
			print("Error: A variogram index column could not be found. Ensure its header is one of"
				  " the\n       following: 'Variogram Index' or 'Index.'")
			return
	if minpairs and experimental:
		if experimental or pairnumbers:
			pairs_name = False
			if 'numpairs' in data.columns:
				pairs_name = 'numpairs'
			if 'Number of Pairs' in data.columns:
				pairs_name = 'Number of Pairs'
			if not pairs_name:
				printerr("The argument `minpairs` is being used; however, the column 'numpairs'"
						 " or 'Number of Pairs' was not found in `data`. `minpairs` has been"
						 " set to a value of `False`.", errtype='warning')
				minpairs = False
		else:
			pairs_name = False
			minpairs = False
			pairnumbers = False
	else:
		pairs_name = False
		minpairs = False
		pairnumbers = False
	# Set-up some default plotting parameters
	if color is None:
		color = Parameters['plotting.variogram_plot.color']
	if experimental:
		if ms is None:
			ms = Parameters['plotting.variogram_plot.ms']
		if marker is None:
			marker = 'o'
		if lw is None:
			lw = 0
		if ls is None and lw > 0:
			ls = '-'
		else:
			ls = 'None'
	else:
		if lw is None:
			lw = 1
		if ls is None:
			ls = '-'
		if ms is None:
			ms = 0
	if label is None:
		if index:
			label = 'Var %s' % index
		else:
			label = 'Var'
	elif not label:
		label = "_nolegend_"
	# Get the right data
	data = data[data[vario_name] != np.nan]
	# This shouldn't be necessary once everything is converted to an nan standard, but for now..
	data = data[data[vario_name] != Parameters['data.null']]
	if experimental:
		data = data[data[h_name] > 0]
	pairs_label = None
	varpairs = None
	if index is None:
		vardata = data[vario_name]
		vardist = data[h_name]
		if minpairs or pairnumbers and experimental:
			varpairs = data[data[pairs_name] <= minpairs]
			pairs_label = data[pairs_name]
	else:
		vardata = data[vario_name][data[index_name] == index]
		vardist = data[h_name][data[index_name] == index]
		if minpairs or pairnumbers and experimental:
			varpairs = data[(data[pairs_name] <= minpairs) & (data[index_name] == index)]
			pairs_label = data[pairs_name]
		if len(vardata) == 0 or len(vardist) == 0:
			print("Error: The index value provided wasn't found")
			return
	# Set-up plot if no axis is supplied
	fig, ax, cax = setup_plot(ax, figsize=figsize, aspect =False)
	# Highlight the pairs with fewer pairs then minpairs if required
	if minpairs:
		for i in range(len(varpairs)):
			try:
				ax.plot(varpairs.iloc[i][h_name], varpairs.iloc[i][vario_name], marker=marker,
						color='#fb8072', ms=round((ms * 1.4 + 0.5), 0))
			except:
				ax.plot(varpairs.iloc[i][h_name], varpairs.iloc[i][vario_name], marker=marker,
						ms=round((ms * 1.4 + 0.5), 0))
	# Plot the variogram and identifier
	ax.plot(vardist, vardata, color=color, lw=lw, marker=marker, ls=ls, ms=ms, label=label,
			**kwargs)
	if experimental and pairnumbers and pairs_label is not None:
		if ylim is None:
			ymin, ymax = ax.get_ybound()
		else:
			ymin, ymax = ylim
		for x, y, lbl in zip(vardist, vardata, pairs_label):
			if ymin <= y <= ymax:
				ax.text(x, y - 0.025, int(lbl), rotation=-55)
	# Plot figure and axis labels
	if unit is None:
		unit = Parameters['plotting.unit']
	if unit is None or unit == '':
		unit = ''
	else:
		unit = ' ({})'.format(unit)
	if xlabel is None:
		xlabel = Parameters['plotting.lagname'] + unit
	if ylabel is None:
		ax.set_ylabel(r'$\gamma$   ', fontsize=mpl.rcParams['font.size'] *
					  Parameters['plotting.gammasize'], rotation=0)
	elif ylabel:
		ax.set_ylabel(ylabel)
	# Draw Sill
	if sill:
		ax.axhline(sill, color='black')
	# Set plot limits
	if ylim is None:
		if sill == 1:
			ylim = (0, 1.2)
		elif abs(sill) < 1:
			# If the sill is bellow 1, it's likely a cross variogram
			if sill < 0:
				# Positive default which also draws a line at 0
				ylim = (-1, 0.2)
				ax.axhline(0, color='black', linestyle='--', lw=0.5, dashes=(2, 2))
			if sill >= 0:
				# Positive defaults
				ylim = (0, 1.2)
	if xlim is None:
		xmin, xmax = ax.get_xbound()
		xlim = [0, xmax]

	ax = format_plot(ax, xlabel=xlabel, title=title, grid=grid, axis_xy=axis_xy, xlim=xlim, ylim=ylim)

	# Export figure
	if output_file or ('pdfpages' in out_kws):
		export_image(output_file, **out_kws)
	# Return axis
	return ax
